@@ -7,19 +7,17 @@ from zope.interface import implements
 from twisted.python import log
 from twisted.internet import defer, reactor
 from twisted.application import service
-import twisted.spread.pb
+from twisted.spread import pb
 
-from buildbot.pbutil import NewCredPerspective
 from buildbot.status.builder import SlaveStatus
 from buildbot.status.mail import MailNotifier
 from buildbot.interfaces import IBuildSlave, ILatentBuildSlave
 from buildbot.process.properties import Properties
+from buildbot.locks import LockAccess
 
 import sys
-if sys.version_info[:3] < (2,4,0):
-    from sets import Set as set
 
-class AbstractBuildSlave(NewCredPerspective, service.MultiService):
+class AbstractBuildSlave(pb.Avatar, service.MultiService):
     """This is the master-side representative for a remote buildbot slave.
     There is exactly one for each slave described in the config file (the
     c['slaves'] list). When buildbots connect in (.attach), they get a
@@ -34,7 +32,7 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
 
     def __init__(self, name, password, max_builds=None,
                  notify_on_missing=[], missing_timeout=3600,
-                 properties={}):
+                 properties={}, locks=None):
         """
         @param name: botname this machine will supply when it connects
         @param password: password this machine will supply when
@@ -45,6 +43,9 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
         @param properties: properties that will be applied to builds run on
                            this slave
         @type properties: dictionary
+        @param locks: A list of locks that must be acquired before this slave
+                      can be used
+        @type locks: dictionary
         """
         service.MultiService.__init__(self)
         self.slavename = name
@@ -55,6 +56,9 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
         self.slave_commands = None
         self.slavebuilders = {}
         self.max_builds = max_builds
+        self.access = []
+        if locks:
+            self.access = locks
 
         self.properties = Properties()
         self.properties.update(properties, "BuildSlave")
@@ -80,6 +84,14 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
         assert self.password == new.password
         assert self.__class__ == new.__class__
         self.max_builds = new.max_builds
+        self.access = new.access
+        self.notify_on_missing = new.notify_on_missing
+        self.missing_timeout = new.missing_timeout
+
+        self.properties.updateFromProperties(new.properties)
+
+        if self.botmaster:
+            self.updateLocks()
 
     def __repr__(self):
         if self.botmaster:
@@ -91,9 +103,56 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
             return "<%s '%s', (no builders yet)>" % \
                 (self.__class__.__name__, self.slavename)
 
+    def updateLocks(self):
+        # convert locks into their real form
+        locks = []
+        for access in self.access:
+            if not isinstance(access, LockAccess):
+                access = access.defaultAccess()
+            lock = self.botmaster.getLockByID(access.lockid)
+            locks.append((lock, access))
+        self.locks = [(l.getLock(self), la) for l, la in locks]
+
+    def locksAvailable(self):
+        """
+        I am called to see if all the locks I depend on are available,
+        in which I return True, otherwise I return False
+        """
+        if not self.locks:
+            return True
+        for lock, access in self.locks:
+            if not lock.isAvailable(access):
+                return False
+        return True
+
+    def acquireLocks(self):
+        """
+        I am called when a build is preparing to run. I try to claim all
+        the locks that are needed for a build to happen. If I can't, then
+        my caller should give up the build and try to get another slave
+        to look at it.
+        """
+        log.msg("acquireLocks(slave %s, locks %s)" % (self, self.locks))
+        if not self.locksAvailable():
+            log.msg("slave %s can't lock, giving up" % (self, ))
+            return False
+        # all locks are available, claim them all
+        for lock, access in self.locks:
+            lock.claim(self, access)
+        return True
+
+    def releaseLocks(self):
+        """
+        I am called to release any locks after a build has finished
+        """
+        log.msg("releaseLocks(%s): %s" % (self, self.locks))
+        for lock, access in self.locks:
+            lock.release(self, access)
+
     def setBotmaster(self, botmaster):
         assert not self.botmaster, "BuildSlave already has a botmaster"
         self.botmaster = botmaster
+        self.updateLocks()
         self.startMissingTimer()
 
     def stopMissingTimer(self):
@@ -107,6 +166,13 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
             self.missing_timer = reactor.callLater(self.missing_timeout,
                                                    self._missing_timer_fired)
 
+    def recordConnectTime(self):
+        if self.slave_status:
+            self.slave_status.recordConnectTime()
+
+    def isConnected(self):
+        return self.slave
+
     def _missing_timer_fired(self):
         self.missing_timer = None
         # notify people, but only if we're still in the config
@@ -151,26 +217,12 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
     def attached(self, bot):
         """This is called when the slave connects.
 
-        @return: a Deferred that fires with a suitable pb.IPerspective to
-                 give to the slave (i.e. 'self')"""
+        @return: a Deferred that fires when the attachment is complete
+        """
 
-        if self.slave:
-            # uh-oh, we've got a duplicate slave. The most likely
-            # explanation is that the slave is behind a slow link, thinks we
-            # went away, and has attempted to reconnect, so we've got two
-            # "connections" from the same slave, but the previous one is
-            # stale. Give the new one precedence.
-            log.msg("duplicate slave %s replacing old one" % self.slavename)
+        # the botmaster should ensure this.
+        assert not self.isConnected()
 
-            # just in case we've got two identically-configured slaves,
-            # report the IP addresses of both so someone can resolve the
-            # squabble
-            tport = self.slave.broker.transport
-            log.msg("old slave was connected from", tport.getPeer())
-            log.msg("new slave is from", bot.broker.transport.getPeer())
-            d = self.disconnect()
-        else:
-            d = defer.succeed(None)
         # now we go through a sequence of calls, gathering information, then
         # tell the Botmaster that it can finally give this slave to all the
         # Builders that care about it.
@@ -184,6 +236,7 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
         # We want to know when the graceful shutdown flag changes
         self.slave_status.addGracefulWatcher(self._gracefulChanged)
 
+        d = defer.succeed(None)
         def _log_attachment_on_slave(res):
             d1 = bot.callRemote("print", "attached")
             d1.addErrback(lambda why: None)
@@ -242,10 +295,11 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
             log.msg("bot attached")
             self.messageReceivedFromSlave()
             self.stopMissingTimer()
+            self.botmaster.parent.status.slaveConnected(self.slavename)
 
             return self.updateSlave()
         d.addCallback(_accept_slave)
-        d.addCallback(lambda res: self.botmaster.maybeStartAllBuilds())
+        d.addCallback(lambda res: self.botmaster.triggerNewBuildCheck())
 
         # Finally, the slave gets a reference to this BuildSlave. They
         # receive this later, after we've started using them.
@@ -262,6 +316,7 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
         self.slave_status.removeGracefulWatcher(self._gracefulChanged)
         self.slave_status.setConnected(False)
         log.msg("BuildSlave.detached(%s)" % self.slavename)
+        self.botmaster.parent.status.slaveDisconnected(self.slavename)
 
     def disconnect(self):
         """Forcibly disconnect the slave.
@@ -316,7 +371,6 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
             # however, these hacks are pretty internal, so don't blow up if
             # they fail or are unavailable
             log.msg("failed to accelerate the shutdown process")
-            pass
         log.msg("waiting for slave to finish disconnecting")
 
         return d
@@ -355,6 +409,10 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
                                if sb.isBusy()]
             if len(active_builders) >= self.max_builds:
                 return False
+
+        if not self.locksAvailable():
+            return False
+
         return True
 
     def _mail_missing_message(self, subject, text):
@@ -411,7 +469,7 @@ class AbstractBuildSlave(NewCredPerspective, service.MultiService):
             # it's because the connection was lost, that means the slave
             # shutdown as expected.
             def _errback(why):
-                if why.check(twisted.spread.pb.PBConnectionLost):
+                if why.check(pb.PBConnectionLost):
                     log.msg("Lost connection to %s" % self.slavename)
                 else:
                     log.err("Unexpected error when trying to shutdown %s" % self.slavename)
@@ -472,15 +530,15 @@ class AbstractLatentBuildSlave(AbstractBuildSlave):
     substantiated = False
     substantiation_deferred = None
     build_wait_timer = None
-    _start_result = _shutdown_callback_handle = None
+    _shutdown_callback_handle = None
 
     def __init__(self, name, password, max_builds=None,
                  notify_on_missing=[], missing_timeout=60*20,
                  build_wait_timeout=60*10,
-                 properties={}):
+                 properties={}, locks=None):
         AbstractBuildSlave.__init__(
             self, name, password, max_builds, notify_on_missing,
-            missing_timeout, properties)
+            missing_timeout, properties, locks)
         self.building = set()
         self.build_wait_timeout = build_wait_timeout
 
@@ -498,7 +556,7 @@ class AbstractLatentBuildSlave(AbstractBuildSlave):
         if self.substantiated:
             self._clearBuildWaitTimer()
             self._setBuildWaitTimer()
-            return defer.succeed(self)
+            return defer.succeed(True)
         if self.substantiation_deferred is None:
             if self.parent and not self.missing_timer:
                 # start timer.  if timer times out, fail deferred
@@ -517,8 +575,13 @@ class AbstractLatentBuildSlave(AbstractBuildSlave):
         d = self.start_instance()
         self._shutdown_callback_handle = reactor.addSystemEventTrigger(
             'before', 'shutdown', self._soft_disconnect, fast=True)
-        def stash_reply(result):
-            self._start_result = result
+        def start_instance_result(result):
+            # If we don't report success, then preparation failed.
+            if not result:
+                log.msg("Slave '%s' doesn not want to substantiate at this time" % (self.slavename,))
+                self.substantiation_deferred.callback(False)
+                self.substantiation_deferred = None
+            return result
         def clean_up(failure):
             if self.missing_timer is not None:
                 self.missing_timer.cancel()
@@ -528,7 +591,7 @@ class AbstractLatentBuildSlave(AbstractBuildSlave):
                 del self._shutdown_callback_handle
                 reactor.removeSystemEventTrigger(handle)
             return failure
-        d.addCallbacks(stash_reply, clean_up)
+        d.addCallbacks(start_instance_result, clean_up)
         return d
 
     def attached(self, bot):
@@ -546,10 +609,11 @@ class AbstractLatentBuildSlave(AbstractBuildSlave):
             self._substantiate()
 
     def _substantiation_failed(self, failure):
-        d = self.substantiation_deferred
-        self.substantiation_deferred = None
         self.missing_timer = None
-        d.errback(failure)
+        if self.substantiation_deferred:
+            d = self.substantiation_deferred
+            self.substantiation_deferred = None
+            d.errback(failure)
         self.insubstantiate()
         # notify people, but only if we're still in the config
         if not self.parent or not self.notify_on_missing:
@@ -685,13 +749,15 @@ class AbstractLatentBuildSlave(AbstractBuildSlave):
             return why
         d.addCallbacks(_sent, _set_failed)
         def _substantiated(res):
+            log.msg("Slave %s substantiated \o/" % self.slavename)
             self.substantiated = True
+            if not self.substantiation_deferred:
+                log.msg("No substantiation deferred for %s" % self.slavename)
             if self.substantiation_deferred:
+                log.msg("Firing %s substantiation deferred with success" % self.slavename)
                 d = self.substantiation_deferred
                 del self.substantiation_deferred
-                res = self._start_result
-                del self._start_result
-                d.callback(res)
+                d.callback(True)
             # note that the missing_timer is already handled within
             # ``attached``
             if not self.building:
