@@ -19,7 +19,6 @@ class ChangePerspective(NewCredPerspective):
     def perspective_addChange(self, changedict):
         log.msg("perspective_addChange called")
         pathnames = []
-        prefixpaths = None
         for path in changedict['files']:
             if self.prefix:
                 if not path.startswith(self.prefix):
@@ -29,13 +28,17 @@ class ChangePerspective(NewCredPerspective):
             pathnames.append(path)
 
         if pathnames:
-            change = changes.Change(changedict['who'],
-                                    pathnames,
-                                    changedict['comments'],
+            change = changes.Change(who=changedict['who'],
+                                    files=pathnames,
+                                    comments=changedict['comments'],
                                     branch=changedict.get('branch'),
                                     revision=changedict.get('revision'),
+                                    revlink=changedict.get('revlink', ''),
                                     category=changedict.get('category'),
                                     when=changedict.get('when'),
+                                    properties=changedict.get('properties', {}),
+                                    repository=changedict.get('repository', '') or '',
+                                    project=changedict.get('project', '') or '',
                                     )
             self.changemaster.addChange(change)
 
