@@ -1,8 +1,10 @@
-
 """Interface documentation.
 
 Define the interfaces that are implemented by various buildbot classes.
 """
+# E0211: Method has no argument
+# E0213: Method should have "self" as first argument
+# pylint: disable-msg=E0211,E0213
 
 from zope.interface import Interface, Attribute
 
@@ -20,6 +22,9 @@ class LatentBuildSlaveFailedToSubstantiate(Exception):
 class BuildbotNotRunningError(Exception):
     pass
 
+class ParameterError(Exception):
+    pass
+
 class IChangeSource(Interface):
     """Object which feeds Change objects to the changemaster. When files or
     directories are changed and the version control system provides some
@@ -103,10 +108,16 @@ class ISourceStamp(Interface):
     @cvar patch: patch applied to the source, or None if no patch
     @type patch: None or tuple (level diff)
 
-    @cvar changes: the source step should check out hte latest revision
+    @cvar changes: the source step should check out the latest revision
                    in the given changes
     @type changes: tuple of L{buildbot.changes.changes.Change} instances,
                    all of which are on the same branch
+
+    @cvar project: project this source code represents
+    @type project: string
+
+    @cvar repository: repository from which source was drawn
+    @type repository: string
     """
 
     def canBeMergedWith(self, other):
@@ -271,7 +282,9 @@ class IBuildSetStatus(Interface):
         """Return a Deferred that fires (with this IBuildSetStatus object)
         when all builds have finished."""
     def getResults():
-        pass
+        """Return SUCCESS/FAILURE, or None if the buildset is not finished
+        yet"""
+
 
 class IBuildRequestStatus(Interface):
     """I represent a request to build a particular set of source code on a
@@ -443,7 +456,7 @@ class IBuilderStatus(Interface):
         delivered."""
 
 class IEventSource(Interface):
-    def eventGenerator(branches=[], categories=[]):
+    def eventGenerator(branches=[], categories=[], committers=[], minTime=0):
         """This function creates a generator which will yield all of this
         object's status events, starting with the most recent and progressing
         backwards in time. These events provide the IStatusEvent interface.
@@ -459,6 +472,13 @@ class IEventSource(Interface):
         should only return events that are categorized within the
         given category.  If the list is empty, events for all
         categories should be returned.
+
+        @param comitters: a list of committers.  The generator should only
+        return events caused by one of the listed committers. If the list is
+        empty or None, events from every committers should be returned.
+
+        @param minTime: a timestamp. Do not generate events occuring prior to
+        this timestamp.
         """
 
 class IBuildStatus(Interface):
@@ -932,6 +952,10 @@ class IStatusReceiver(Interface):
         """This is a periodic update on the progress this Build has made
         towards completion."""
 
+    def changeAdded(change):
+        """A new Change was added to the ChangeMaster.  By the time this event
+        is received, all schedulers have already received the change."""
+
     def stepStarted(build, step):
         """A step has just started. 'step' is the IBuildStepStatus which
         represents the step: it can be queried for more information.
@@ -1000,30 +1024,39 @@ class IStatusReceiver(Interface):
     def builderRemoved(builderName):
         """The Builder has been removed."""
 
+    def slaveConnected(slaveName):
+        """The slave has connected."""
+
+    def slaveDisconnected(slaveName):
+        """The slave has disconnected."""
+
 class IControl(Interface):
     def addChange(change):
         """Add a change to all builders. Each Builder will decide for
         themselves whether the change is interesting or not, and may initiate
         a build as a result."""
 
-    def submitBuildSet(buildset):
-        """Submit a BuildSet object, which will eventually be run on all of
-        the builders listed therein."""
+    def submitBuildSet(builderNames, ss, reason, props=None, now=False):
+        """Create a BuildSet, which will eventually cause a build of the
+        given SourceStamp to be run on all of the named builders. This
+        returns a BuildSetStatus object, which can be used to keep track of
+        the builds that are performed.
+
+        If now=True, and the builder has no slave attached, NoSlaveError will
+        be raised instead of queueing the request for later action."""
 
     def getBuilder(name):
         """Retrieve the IBuilderControl object for the given Builder."""
 
 class IBuilderControl(Interface):
-    def requestBuild(request):
-        """Queue a L{buildbot.process.base.BuildRequest} object for later
-        building."""
+    def submitBuildRequest(ss, reason, props=None, now=False):
+        """Create a BuildRequest, which will eventually cause a build of the
+        given SourceStamp to be run on this builder. This returns a
+        BuildRequestStatus object, which can be used to keep track of the
+        builds that are performed.
 
-    def requestBuildSoon(request):
-        """Submit a BuildRequest like requestBuild, but raise a
-        L{buildbot.interfaces.NoSlaveError} if no slaves are currently
-        available, so it cannot be used to queue a BuildRequest in the hopes
-        that a slave will eventually connect. This method is appropriate for
-        use by things like the web-page 'Force Build' button."""
+        If now=True, and I have no slave attached, NoSlaveError will be
+        raised instead of queueing the request for later action."""
 
     def resubmitBuild(buildStatus, reason="<rebuild, no reason given>"):
         """Rebuild something we've already built before. This submits a
@@ -1043,12 +1076,13 @@ class IBuilderControl(Interface):
         the build that is currently in progress: once the build finishes,
         there is nothing to control anymore."""
 
-    def ping(timeout=30):
+    def ping():
         """Attempt to contact the slave and see if it is still alive. This
         returns a Deferred which fires with either True (the slave is still
-        alive) or False (the slave did not respond). As a side effect, adds
-        an event to this builder's column in the waterfall display
-        containing the results of the ping."""
+        alive) or False (the slave did not respond). As a side effect, adds an
+        event to this builder's column in the waterfall display containing the
+        results of the ping. Note that this may not fail for a long time, it is
+        implemented in terms of the timeout on the underlying TCP connection."""
         # TODO: this ought to live in ISlaveControl, maybe with disconnect()
         # or something. However the event that is emitted is most useful in
         # the Builder column, so it kinda fits here too.
