@@ -8,144 +8,10 @@ from twisted.python.failure import Failure
 from twisted.internet import reactor, defer, error
 
 from buildbot import interfaces, locks
-from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, EXCEPTION
-from buildbot.status.builder import Results, BuildRequestStatus
+from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, EXCEPTION, \
+  RETRY, worst_status
+from buildbot.status.builder import Results
 from buildbot.status.progress import BuildProgress
-from buildbot.process.properties import Properties
-
-class BuildRequest:
-    """I represent a request to a specific Builder to run a single build.
-
-    I have a SourceStamp which specifies what sources I will build. This may
-    specify a specific revision of the source tree (so source.branch,
-    source.revision, and source.patch are used). The .patch attribute is
-    either None or a tuple of (patchlevel, diff), consisting of a number to
-    use in 'patch -pN', and a unified-format context diff.
-
-    Alternatively, the SourceStamp may specify a set of Changes to be built,
-    contained in source.changes. In this case, I may be mergeable with other
-    BuildRequests on the same branch.
-
-    I may be part of a BuildSet, in which case I will report status results
-    to it.
-
-    I am paired with a BuildRequestStatus object, to which I feed status
-    information.
-
-    @type source: a L{buildbot.sourcestamp.SourceStamp} instance.   
-    @ivar source: the source code that this BuildRequest use
-
-    @type reason: string
-    @ivar reason: the reason this Build is being requested. Schedulers
-                  provide this, but for forced builds the user requesting the
-                  build will provide a string.
-
-    @type properties: Properties object
-    @ivar properties: properties that should be applied to this build
-                      'owner' property is used by Build objects to collect
-                      the list returned by getInterestedUsers
-
-    @ivar status: the IBuildStatus object which tracks our status
-
-    @ivar submittedAt: a timestamp (seconds since epoch) when this request
-                       was submitted to the Builder. This is used by the CVS
-                       step to compute a checkout timestamp, as well as the
-                       master to prioritize build requests from oldest to
-                       newest.
-    """
-
-    source = None
-    builder = None
-    startCount = 0 # how many times we have tried to start this build
-    submittedAt = None
-
-    implements(interfaces.IBuildRequestControl)
-
-    def __init__(self, reason, source, builderName, properties=None):
-        assert interfaces.ISourceStamp(source, None)
-        self.reason = reason
-        self.source = source
-
-        self.properties = Properties()
-        if properties:
-            self.properties.updateFromProperties(properties)
-
-        self.start_watchers = []
-        self.finish_watchers = []
-        self.status = BuildRequestStatus(source, builderName)
-
-    def canBeMergedWith(self, other):
-        return self.source.canBeMergedWith(other.source)
-
-    def mergeWith(self, others):
-        return self.source.mergeWith([o.source for o in others])
-
-    def mergeReasons(self, others):
-        """Return a reason for the merged build request."""
-        reasons = []
-        for req in [self] + others:
-            if req.reason and req.reason not in reasons:
-                reasons.append(req.reason)
-        return ", ".join(reasons)
-
-    def waitUntilFinished(self):
-        """Get a Deferred that will fire (with a
-        L{buildbot.interfaces.IBuildStatus} instance when the build
-        finishes."""
-        d = defer.Deferred()
-        self.finish_watchers.append(d)
-        return d
-
-    # these are called by the Builder
-
-    def requestSubmitted(self, builder):
-        # the request has been placed on the queue
-        self.builder = builder
-
-    def buildStarted(self, build, buildstatus):
-        """This is called by the Builder when a Build has been started in the
-        hopes of satifying this BuildRequest. It may be called multiple
-        times, since interrupted builds and lost buildslaves may force
-        multiple Builds to be run until the fate of the BuildRequest is known
-        for certain."""
-        for o in self.start_watchers[:]:
-            # these observers get the IBuildControl
-            o(build)
-        # while these get the IBuildStatus
-        self.status.buildStarted(buildstatus)
-
-    def finished(self, buildstatus):
-        """This is called by the Builder when the BuildRequest has been
-        retired. This happens when its Build has either succeeded (yay!) or
-        failed (boo!). TODO: If it is halted due to an exception (oops!), or
-        some other retryable error, C{finished} will not be called yet."""
-
-        for w in self.finish_watchers:
-            w.callback(buildstatus)
-        self.finish_watchers = []
-
-    # IBuildRequestControl
-
-    def subscribe(self, observer):
-        self.start_watchers.append(observer)
-    def unsubscribe(self, observer):
-        self.start_watchers.remove(observer)
-
-    def cancel(self):
-        """Cancel this request. This can only be successful if the Build has
-        not yet been started.
-
-        @return: a boolean indicating if the cancel was successful."""
-        if self.builder:
-            return self.builder.cancelBuildRequest(self)
-        return False
-
-    def setSubmitTime(self, t):
-        self.submittedAt = t
-        self.status.setSubmitTime(t)
-
-    def getSubmitTime(self):
-        return self.submittedAt
 
 
 class Build:
@@ -178,6 +44,7 @@ class Build:
     reason = "changes"
     finished = False
     results = None
+    stopped = False
 
     def __init__(self, requests):
         self.requests = requests
@@ -194,6 +61,8 @@ class Build:
 
         self.terminate = False
 
+        self._acquiringLock = None
+
     def setBuilder(self, builder):
         """
         Set the given builder as our builder.
@@ -211,11 +80,11 @@ class Build:
     def getSourceStamp(self):
         return self.source
 
-    def setProperty(self, propname, value, source):
+    def setProperty(self, propname, value, source, runtime=True):
         """Set a property on this build. This may only be called after the
         build has started, so that it has a BuildStatus object where the
         properties can live."""
-        self.build_status.setProperty(propname, value, source)
+        self.build_status.setProperty(propname, value, source, runtime=True)
 
     def getProperties(self):
         return self.build_status.getProperties()
@@ -229,12 +98,9 @@ class Build:
     def allFiles(self):
         # return a list of all source files that were changed
         files = []
-        havedirs = 0
         for c in self.allChanges():
             for f in c.files:
                 files.append(f)
-            if c.isdir:
-                havedirs = 1
         return files
 
     def __repr__(self):
@@ -280,21 +146,23 @@ class Build:
         buildmaster = self.builder.botmaster.parent
         props.updateFromProperties(buildmaster.properties)
 
-        # get any properties from requests (this is the path through
-        # which schedulers will send us properties)
-        for rq in self.requests:
-            props.updateFromProperties(rq.properties)
-
-        # and finally, from the SourceStamp, which has properties via Change
+        # from the SourceStamp, which has properties via Change
         for change in self.source.changes:
             props.updateFromProperties(change.properties)
 
+        # and finally, get any properties from requests (this is the path
+        # through which schedulers will send us properties)
+        for rq in self.requests:
+            props.updateFromProperties(rq.properties)
+
         # now set some properties of our own, corresponding to the
         # build itself
-        props.setProperty("buildername", self.builder.name, "Build")
         props.setProperty("buildnumber", self.build_status.number, "Build")
         props.setProperty("branch", self.source.branch, "Build")
         props.setProperty("revision", self.source.revision, "Build")
+        props.setProperty("repository", self.source.repository, "Build")
+        props.setProperty("project", self.source.project, "Build")
+        self.builder.setupProperties(props)
 
     def setupSlaveBuilder(self, slavebuilder):
         self.slavebuilder = slavebuilder
@@ -349,8 +217,9 @@ class Build:
             self.setupBuild(expectations) # create .steps
         except:
             # the build hasn't started yet, so log the exception as a point
-            # event instead of flunking the build. TODO: associate this
-            # failure with the build instead. this involves doing
+            # event instead of flunking the build. 
+            # TODO: associate this failure with the build instead. 
+            # this involves doing
             # self.build_status.buildStarted() from within the exception
             # handler
             log.msg("Build.setupBuild failed")
@@ -363,18 +232,23 @@ class Build:
             d.callback(self)
             return d
 
+        self.build_status.buildStarted(self)
         self.acquireLocks().addCallback(self._startBuild_2)
         return d
 
     def acquireLocks(self, res=None):
-        log.msg("acquireLocks(step %s, locks %s)" % (self, self.locks))
+        self._acquiringLock = None
         if not self.locks:
             return defer.succeed(None)
+        if self.stopped:
+            return defer.succeed(None)
+        log.msg("acquireLocks(build %s, locks %s)" % (self, self.locks))
         for lock, access in self.locks:
             if not lock.isAvailable(access):
                 log.msg("Build %s waiting for lock %s" % (self, lock))
                 d = lock.waitUntilMaybeAvailable(self, access)
                 d.addCallback(self.acquireLocks)
+                self._acquiringLock = (lock, access, d)
                 return d
         # all locks are available, claim them all
         for lock, access in self.locks:
@@ -382,7 +256,6 @@ class Build:
         return defer.succeed(None)
 
     def _startBuild_2(self, res):
-        self.build_status.buildStarted(self)
         self.startNextStep()
 
     def setupBuild(self, expectations):
@@ -403,7 +276,10 @@ class Build:
                 raise
             step.setBuild(self)
             step.setBuildSlave(self.slavebuilder.slave)
-            step.setDefaultWorkdir(self.workdir)
+            if callable (self.workdir):
+                step.setDefaultWorkdir (self.workdir (self.source))
+            else:
+                step.setDefaultWorkdir (self.workdir)
             name = step.name
             if stepnames.has_key(name):
                 count = stepnames[name]
@@ -443,7 +319,6 @@ class Build:
 
         # we are now ready to set up our BuildStatus.
         self.build_status.setSourceStamp(self.source)
-        self.build_status.setRequests([req.status for req in self.requests])
         self.build_status.setReason(self.reason)
         self.build_status.setBlamelist(self.blamelist())
         self.build_status.setProgress(self.progress)
@@ -461,8 +336,12 @@ class Build:
         """This method is called to obtain the next BuildStep for this build.
         When it returns None (or raises a StopIteration exception), the build
         is complete."""
+        if self.stopped:
+            return None
         if not self.steps:
             return None
+        if not self.remote:
+            return None
         if self.terminate:
             while True:
                 s = self.steps.pop(0)
@@ -510,23 +389,28 @@ class Build:
             self.text.extend(text)
         if not self.remote:
             terminate = True
+
+        possible_overall_result = result
         if result == FAILURE:
+            if not step.flunkOnFailure:
+                possible_overall_result = SUCCESS
             if step.warnOnFailure:
-                if self.result != FAILURE:
-                    self.result = WARNINGS
+                possible_overall_result = WARNINGS
             if step.flunkOnFailure:
-                self.result = FAILURE
+                possible_overall_result = FAILURE
             if step.haltOnFailure:
                 terminate = True
         elif result == WARNINGS:
-            if step.warnOnWarnings:
-                if self.result != FAILURE:
-                    self.result = WARNINGS
+            if not step.warnOnWarnings:
+                possible_overall_result = SUCCESS
+            else:
+                possible_overall_result = WARNINGS
             if step.flunkOnWarnings:
-                self.result = FAILURE
-        elif result == EXCEPTION:
-            self.result = EXCEPTION
+                possible_overall_result = FAILURE
+        elif result in (EXCEPTION, RETRY):
             terminate = True
+
+        self.result = worst_status(self.result, possible_overall_result)
         return terminate
 
     def lostRemote(self, remote=None):
@@ -553,14 +437,17 @@ class Build:
             return
         # TODO: include 'reason' in this point event
         self.builder.builder_status.addPointEvent(['interrupt'])
-        self.currentStep.interrupt(reason)
-        if 0:
-            # TODO: maybe let its deferred do buildFinished
-            if self.currentStep and self.currentStep.progress:
-                # XXX: really .fail or something
-                self.currentStep.progress.finish()
-            text = ["stopped", reason]
-            self.buildFinished(text, FAILURE)
+        self.stopped = True
+        if self.currentStep:
+            self.currentStep.interrupt(reason)
+
+        self.result = FAILURE
+        self.text.append("Interrupted")
+
+        if self._acquiringLock:
+            lock, access, d = self._acquiringLock
+            lock.stopWaitingUntilAvailable(self, access, d)
+            d.callback(None)
 
     def allStepsDone(self):
         if self.result == FAILURE:
@@ -609,9 +496,14 @@ class Build:
         self.deferred = None
 
     def releaseLocks(self):
-        log.msg("releaseLocks(%s): %s" % (self, self.locks))
+        if self.locks:
+            log.msg("releaseLocks(%s): %s" % (self, self.locks))
         for lock, access in self.locks:
-            lock.release(self, access)
+            if lock.isOwner(self, access):
+                lock.release(self, access)
+            else:
+                # This should only happen if we've been interrupted
+                assert self.stopped
 
     # IBuildControl
 
