@@ -1,21 +1,25 @@
 
 import random, weakref
 from zope.interface import implements
-from twisted.python import log, components
+from twisted.python import log
+from twisted.python.failure import Failure
 from twisted.spread import pb
-from twisted.internet import reactor, defer
+from twisted.application import service, internet
+from twisted.internet import defer
 
-from buildbot import interfaces
+from buildbot import interfaces, util
 from buildbot.status.progress import Expectations
-from buildbot.util import now
-from buildbot.process import base
+from buildbot.status.builder import RETRY
+from buildbot.process.properties import Properties
+from buildbot.util.eventual import eventually
 
 (ATTACHING, # slave attached, still checking hostinfo/etc
  IDLE, # idle, available for use
  PINGING, # build about to start, making sure it is still alive
  BUILDING, # build is running
  LATENT, # latent slave is not substantiated; similar to idle
- ) = range(5)
+ SUBSTANTIATING,
+ ) = range(6)
 
 
 class AbstractSlaveBuilder(pb.Referenceable):
@@ -30,13 +34,14 @@ class AbstractSlaveBuilder(pb.Referenceable):
         self.remote = None
         self.slave = None
         self.builder_name = None
+        self.locks = None
 
     def __repr__(self):
         r = ["<", self.__class__.__name__]
         if self.builder_name:
-            r.extend([" builder=", self.builder_name])
+            r.extend([" builder=", repr(self.builder_name)])
         if self.slave:
-            r.extend([" slave=", self.slave.slavename])
+            r.extend([" slave=", repr(self.slave.slavename)])
         r.append(">")
         return ''.join(r)
 
@@ -70,7 +75,7 @@ class AbstractSlaveBuilder(pb.Referenceable):
 
     def buildFinished(self):
         self.state = IDLE
-        reactor.callLater(0, self.builder.botmaster.maybeStartAllBuilds)
+        self.builder.triggerNewBuildCheck()
 
     def attached(self, slave, remote, commands):
         """
@@ -92,29 +97,34 @@ class AbstractSlaveBuilder(pb.Referenceable):
             assert self.slave == slave
         log.msg("Buildslave %s attached to %s" % (slave.slavename,
                                                   self.builder_name))
-        d = self.remote.callRemote("setMaster", self)
-        d.addErrback(self._attachFailure, "Builder.setMaster")
-        d.addCallback(self._attached2)
+        def _attachFailure(why, where):
+            log.msg(where)
+            log.err(why)
+            return why
+
+        d = defer.succeed(None)
+        def doSetMaster(res):
+            d = self.remote.callRemote("setMaster", self)
+            #d.addErrback(_attachFailure, "Builder.setMaster")
+            return d
+        d.addCallback(doSetMaster)
+        def doPrint(res):
+            d = self.remote.callRemote("print", "attached")
+            #d.addErrback(_attachFailure, "Builder.print 'attached'")
+            return d
+        d.addCallback(doPrint)
+        def setIdle(res):
+            self.state = IDLE
+            return self
+        d.addCallback(setIdle)
         return d
 
-    def _attached2(self, res):
-        d = self.remote.callRemote("print", "attached")
-        d.addErrback(self._attachFailure, "Builder.print 'attached'")
-        d.addCallback(self._attached3)
-        return d
+    def prepare(self, builder_status):
+        if not self.slave.acquireLocks():
+            return defer.succeed(False)
+        return defer.succeed(True)
 
-    def _attached3(self, res):
-        # now we say they're really attached
-        self.state = IDLE
-        return self
-
-    def _attachFailure(self, why, where):
-        assert isinstance(where, str)
-        log.msg(where)
-        log.err(why)
-        return why
-
-    def ping(self, timeout, status=None):
+    def ping(self, status=None):
         """Ping the slave to make sure it is still there. Returns a Deferred
         that fires with True if it is.
 
@@ -134,7 +144,7 @@ class AbstractSlaveBuilder(pb.Referenceable):
                 self.ping_watchers.insert(0, d2)
                 # I think it will make the tests run smoother if the status
                 # is updated before the ping completes
-            Ping().ping(self.remote, timeout).addCallback(self._pong)
+            Ping().ping(self.remote).addCallback(self._pong)
 
         def reset_state(res):
             if self.state == PINGING:
@@ -167,10 +177,12 @@ class AbstractSlaveBuilder(pb.Referenceable):
 
 class Ping:
     running = False
-    timer = None
 
-    def ping(self, remote, timeout):
+    def ping(self, remote):
         assert not self.running
+        if not remote:
+            # clearly the ping must fail
+            return defer.succeed(False)
         self.running = True
         log.msg("sending ping")
         self.d = defer.Deferred()
@@ -179,40 +191,14 @@ class Ping:
         remote.callRemote("print", "ping").addCallbacks(self._pong,
                                                         self._ping_failed,
                                                         errbackArgs=(remote,))
-
-        # We use either our own timeout or the (long) TCP timeout to detect
-        # silently-missing slaves. This might happen because of a NAT
-        # timeout or a routing loop. If the slave just shuts down (and we
-        # somehow missed the FIN), we should get a "connection refused"
-        # message.
-        self.timer = reactor.callLater(timeout, self._ping_timeout, remote)
         return self.d
 
-    def _ping_timeout(self, remote):
-        log.msg("ping timeout")
-        # force the BuildSlave to disconnect, since this indicates that
-        # the bot is unreachable.
-        del self.timer
-        remote.broker.transport.loseConnection()
-        # the forcibly-lost connection will now cause the ping to fail
-
-    def _stopTimer(self):
-        if not self.running:
-            return
-        self.running = False
-
-        if self.timer:
-            self.timer.cancel()
-            del self.timer
-
     def _pong(self, res):
         log.msg("ping finished: success")
-        self._stopTimer()
         self.d.callback(True)
 
     def _ping_failed(self, res, remote):
         log.msg("ping finished: failure")
-        self._stopTimer()
         # the slave has some sort of internal error, disconnect them. If we
         # don't, we'll requeue a build and ping them again right away,
         # creating a nasty loop.
@@ -243,9 +229,9 @@ class SlaveBuilder(AbstractSlaveBuilder):
         self.state = IDLE
         if self.slave:
             d = self.slave.buildFinished(self)
-            d.addCallback(lambda x: reactor.callLater(0, self.builder.botmaster.maybeStartAllBuilds))
+            d.addCallback(lambda x: self.builder.triggerNewBuildCheck())
         else:
-            reactor.callLater(0, self.builder.botmaster.maybeStartAllBuilds)
+            self.builder.triggerNewBuildCheck()
 
 
 class LatentSlaveBuilder(AbstractSlaveBuilder):
@@ -258,7 +244,30 @@ class LatentSlaveBuilder(AbstractSlaveBuilder):
         log.msg("Latent buildslave %s attached to %s" % (slave.slavename,
                                                          self.builder_name))
 
-    def substantiate(self, build):
+    def prepare(self, builder_status):
+        # If we can't lock, then don't bother trying to substantiate
+        if not self.slave.acquireLocks():
+            return defer.succeed(False)
+
+        log.msg("substantiating slave %s" % (self,))
+        d = self.substantiate()
+        def substantiation_failed(f):
+            builder_status.addPointEvent(['removing', 'latent',
+                                          self.slave.slavename])
+            self.slave.disconnect()
+            # TODO: should failover to a new Build
+            return f
+        def substantiation_cancelled(res):
+            # if res is False, latent slave cancelled subtantiation
+            if not res:
+                self.state = LATENT
+            return res
+        d.addCallback(substantiation_cancelled)
+        d.addErrback(substantiation_failed)
+        return d
+
+    def substantiate(self):
+        self.state = SUBSTANTIATING
         d = self.slave.substantiate(self)
         if not self.slave.substantiated:
             event = self.builder.builder_status.addEvent(
@@ -296,15 +305,15 @@ class LatentSlaveBuilder(AbstractSlaveBuilder):
         self.state = LATENT
         return AbstractSlaveBuilder._attachFailure(self, why, where)
 
-    def ping(self, timeout, status=None):
+    def ping(self, status=None):
         if not self.slave.substantiated:
             if status:
                 status.addEvent(["ping", "latent"]).finish()
             return defer.succeed(True)
-        return AbstractSlaveBuilder.ping(self, timeout, status)
+        return AbstractSlaveBuilder.ping(self, status)
 
 
-class Builder(pb.Referenceable):
+class Builder(pb.Referenceable, service.MultiService):
     """I manage all Builds of a given type.
 
     Each Builder is created by an entry in the config file (the c['builders']
@@ -332,11 +341,6 @@ class Builder(pb.Referenceable):
     I also manage forced builds, progress expectation (ETA) management, and
     some status delivery chores.
 
-    I am persisted in C{BASEDIR/BUILDERNAME/builder}, so I can remember how
-    long a build usually takes to run (in my C{expectations} attribute). This
-    pickle also includes the L{buildbot.status.builder.BuilderStatus} object,
-    which remembers the set of historic builds.
-
     @type buildable: list of L{buildbot.process.base.BuildRequest}
     @ivar buildable: BuildRequests that are ready to build, but which are
                      waiting for a buildslave to be available.
@@ -349,7 +353,6 @@ class Builder(pb.Referenceable):
     """
 
     expectations = None # this is created the first time we get a good build
-    START_BUILD_TIMEOUT = 10
     CHOOSE_SLAVES_RANDOMLY = True # disabled for determinism during tests
 
     def __init__(self, setup, builder_status):
@@ -357,9 +360,10 @@ class Builder(pb.Referenceable):
         @type  setup: dict
         @param setup: builder setup data, as stored in
                       BuildmasterConfig['builders'].  Contains name,
-                      slavename(s), builddir, factory, locks.
+                      slavename(s), builddir, slavebuilddir, factory, locks.
         @type  builder_status: L{buildbot.status.builder.BuilderStatus}
         """
+        service.MultiService.__init__(self)
         self.name = setup['name']
         self.slavenames = []
         if setup.has_key('slavename'):
@@ -367,7 +371,11 @@ class Builder(pb.Referenceable):
         if setup.has_key('slavenames'):
             self.slavenames.extend(setup['slavenames'])
         self.builddir = setup['builddir']
+        self.slavebuilddir = setup['slavebuilddir']
         self.buildFactory = setup['factory']
+        self.nextSlave = setup.get('nextSlave')
+        if self.nextSlave is not None and not callable(self.nextSlave):
+            raise ValueError("nextSlave must be callable")
         self.locks = setup.get("locks", [])
         self.env = setup.get('env', {})
         assert isinstance(self.env, dict)
@@ -375,9 +383,16 @@ class Builder(pb.Referenceable):
             raise ValueError("periodicBuildTime can no longer be defined as"
                              " part of the Builder: use scheduler.Periodic"
                              " instead")
+        self.nextBuild = setup.get('nextBuild')
+        if self.nextBuild is not None and not callable(self.nextBuild):
+            raise ValueError("nextBuild must be callable")
+        self.buildHorizon = setup.get('buildHorizon')
+        self.logHorizon = setup.get('logHorizon')
+        self.eventHorizon = setup.get('eventHorizon')
+        self.mergeRequests = setup.get('mergeRequests', True)
+        self.properties = setup.get('properties', {})
 
         # build/wannabuild slots: Build objects move along this sequence
-        self.buildable = []
         self.building = []
         # old_building holds active builds that were stolen from a predecessor
         self.old_building = weakref.WeakKeyDictionary()
@@ -393,13 +408,22 @@ class Builder(pb.Referenceable):
 
         self.builder_status = builder_status
         self.builder_status.setSlavenames(self.slavenames)
+        self.builder_status.buildHorizon = self.buildHorizon
+        self.builder_status.logHorizon = self.logHorizon
+        self.builder_status.eventHorizon = self.eventHorizon
+        t = internet.TimerService(10*60, self.reclaimAllBuilds)
+        t.setServiceParent(self)
 
         # for testing, to help synchronize tests
         self.watchers = {'attach': [], 'detach': [], 'detach_all': [],
                          'idle': []}
+        self.run_count = 0
 
     def setBotmaster(self, botmaster):
         self.botmaster = botmaster
+        self.db = botmaster.db
+        self.master_name = botmaster.master_name
+        self.master_incarnation = botmaster.master_incarnation
 
     def compareToSetup(self, setup):
         diffs = []
@@ -413,53 +437,166 @@ class Builder(pb.Referenceable):
         if setup['builddir'] != self.builddir:
             diffs.append('builddir changed from %s to %s' \
                          % (self.builddir, setup['builddir']))
+        if setup['slavebuilddir'] != self.slavebuilddir:
+            diffs.append('slavebuilddir changed from %s to %s' \
+                         % (self.slavebuilddir, setup['slavebuilddir']))
         if setup['factory'] != self.buildFactory: # compare objects
             diffs.append('factory changed')
-        oldlocks = [(lock.__class__, lock.name)
-                    for lock in self.locks]
-        newlocks = [(lock.__class__, lock.name)
-                    for lock in setup.get('locks',[])]
-        if oldlocks != newlocks:
-            diffs.append('locks changed from %s to %s' % (oldlocks, newlocks))
+        if setup.get('locks', []) != self.locks:
+            diffs.append('locks changed from %s to %s' % (self.locks, setup.get('locks')))
+        if setup.get('env', {}) != self.env:
+            diffs.append('env changed from %s to %s' % (self.env, setup.get('env', {})))
+        if setup.get('nextSlave') != self.nextSlave:
+            diffs.append('nextSlave changed from %s to %s' % (self.nextSlave, setup.get('nextSlave')))
+        if setup.get('nextBuild') != self.nextBuild:
+            diffs.append('nextBuild changed from %s to %s' % (self.nextBuild, setup.get('nextBuild')))
+        if setup['buildHorizon'] != self.buildHorizon:
+            diffs.append('buildHorizon changed from %s to %s' % (self.buildHorizon, setup['buildHorizon']))
+        if setup['logHorizon'] != self.logHorizon:
+            diffs.append('logHorizon changed from %s to %s' % (self.logHorizon, setup['logHorizon']))
+        if setup['eventHorizon'] != self.eventHorizon:
+            diffs.append('eventHorizon changed from %s to %s' % (self.eventHorizon, setup['eventHorizon']))
         return diffs
 
     def __repr__(self):
-        return "<Builder '%s' at %d>" % (self.name, id(self))
+        return "<Builder '%r' at %d>" % (self.name, id(self))
+
+    def triggerNewBuildCheck(self):
+        self.botmaster.triggerNewBuildCheck()
+
+    def run(self):
+        """Check for work to be done. This should be called any time I might
+        be able to start a job:
+
+         - when the Builder is first created
+         - when a new job has been added to the [buildrequests] DB table
+         - when a slave has connected
+
+        If I have both an available slave and the database contains a
+        BuildRequest that I can handle, I will claim the BuildRequest and
+        start the build. When the build finishes, I will retire the
+        BuildRequest.
+        """
+        # overall plan:
+        #  move .expectations to DB
+
+        # if we're not running, we may still be called from leftovers from
+        # a run of the loop, so just ignore the call.
+        if not self.running:
+            return
+
+        self.run_count += 1
+
+        available_slaves = [sb for sb in self.slaves if sb.isAvailable()]
+        if not available_slaves:
+            self.updateBigStatus()
+            return
+        d = self.db.runInteraction(self._claim_buildreqs, available_slaves)
+        d.addCallback(self._start_builds)
+        return d
+
+    # slave-managers must refresh their claim on a build at least once an
+    # hour, less any inter-manager clock skew
+    RECLAIM_INTERVAL = 1*3600
+
+    def _claim_buildreqs(self, t, available_slaves):
+        # return a dict mapping slave -> (brid,ssid)
+        now = util.now()
+        old = now - self.RECLAIM_INTERVAL
+        requests = self.db.get_unclaimed_buildrequests(self.name, old,
+                                                       self.master_name,
+                                                       self.master_incarnation,
+                                                       t)
+
+        assignments = {}
+        while requests and available_slaves:
+            sb = self._choose_slave(available_slaves)
+            if not sb:
+                log.msg("%s: want to start build, but we don't have a remote"
+                        % self)
+                break
+            available_slaves.remove(sb)
+            breq = self._choose_build(requests)
+            if not breq:
+                log.msg("%s: went to start build, but nextBuild said not to"
+                        % self)
+                break
+            requests.remove(breq)
+            merged_requests = [breq]
+            for other_breq in requests[:]:
+                if (self.mergeRequests and
+                    self.botmaster.shouldMergeRequests(self, breq, other_breq)
+                    ):
+                    requests.remove(other_breq)
+                    merged_requests.append(other_breq)
+            assignments[sb] = merged_requests
+            brids = [br.id for br in merged_requests]
+            self.db.claim_buildrequests(now, self.master_name,
+                                        self.master_incarnation, brids, t)
+        return assignments
+
+    def _choose_slave(self, available_slaves):
+        # note: this might return None if the nextSlave() function decided to
+        # not give us anything
+        if self.nextSlave:
+            try:
+                return self.nextSlave(self, available_slaves)
+            except:
+                log.msg("Exception choosing next slave")
+                log.err(Failure())
+            return None
+        if self.CHOOSE_SLAVES_RANDOMLY:
+            return random.choice(available_slaves)
+        return available_slaves[0]
+
+    def _choose_build(self, buildable):
+        if self.nextBuild:
+            try:
+                return self.nextBuild(self, buildable)
+            except:
+                log.msg("Exception choosing next build")
+                log.err(Failure())
+            return None
+        return buildable[0]
+
+    def _start_builds(self, assignments):
+        # because _claim_buildreqs runs in a separate thread, we might have
+        # lost a slave by this point. We treat that case the same as if we
+        # lose the slave right after the build starts: the initial ping
+        # fails.
+        for (sb, requests) in assignments.items():
+            build = self.buildFactory.newBuild(requests)
+            build.setBuilder(self)
+            build.setLocks(self.locks)
+            if len(self.env) > 0:
+                build.setSlaveEnvironment(self.env)
+            self.startBuild(build, sb)
+        self.updateBigStatus()
+
+
+    def getBuildable(self, limit=None):
+        return self.db.runInteractionNow(self._getBuildable, limit)
+    def _getBuildable(self, t, limit):
+        now = util.now()
+        old = now - self.RECLAIM_INTERVAL
+        return self.db.get_unclaimed_buildrequests(self.name, old,
+                                                   self.master_name,
+                                                   self.master_incarnation,
+                                                   t,
+                                                   limit)
 
     def getOldestRequestTime(self):
         """Returns the timestamp of the oldest build request for this builder.
 
         If there are no build requests, None is returned."""
-        if self.buildable:
-            return self.buildable[0].getSubmitTime()
-        else:
-            return None
+        buildable = self.getBuildable(1)
+        if buildable:
+            # TODO: this is sorted by priority first, not strictly reqtime
+            return buildable[0].getSubmitTime()
+        return None
 
-    def submitBuildRequest(self, req):
-        req.setSubmitTime(now())
-        self.buildable.append(req)
-        req.requestSubmitted(self)
-        self.builder_status.addBuildRequest(req.status)
-        self.maybeStartBuild()
-
-    def cancelBuildRequest(self, req):
-        if req in self.buildable:
-            self.buildable.remove(req)
-            self.builder_status.removeBuildRequest(req.status)
-            return True
-        return False
-
-    def __getstate__(self):
-        d = self.__dict__.copy()
-        # TODO: note that d['buildable'] can contain Deferreds
-        del d['building'] # TODO: move these back to .buildable?
-        del d['slaves']
-        return d
-
-    def __setstate__(self, d):
-        self.__dict__ = d
-        self.building = []
-        self.slaves = []
+    def cancelBuildRequest(self, brid):
+        return self.db.cancel_buildrequests([brid])
 
     def consumeTheSoulOfYourPredecessor(self, old):
         """Suck the brain out of an old Builder.
@@ -474,12 +611,9 @@ class Builder(pb.Referenceable):
 
         log.msg("consumeTheSoulOfYourPredecessor: %s feeding upon %s" %
                 (self, old))
-        # we claim all the pending builds, removing them from the old
-        # Builder's queue. This insures that the old Builder will not start
-        # any new work.
-        log.msg(" stealing %s buildrequests" % len(old.buildable))
-        self.buildable.extend(old.buildable)
-        old.buildable = []
+        # all pending builds are stored in the DB, so we don't have to do
+        # anything to claim them. The old builder will be stopService'd,
+        # which should make sure they don't start any new work
 
         # old.building (i.e. builds which are still running) is not migrated
         # directly: it keeps track of builds which were in progress in the
@@ -527,12 +661,22 @@ class Builder(pb.Referenceable):
 
         return # all done
 
+    def reclaimAllBuilds(self):
+        now = util.now()
+        brids = set()
+        for b in self.building:
+            brids.update([br.id for br in b.requests])
+        for b in self.old_building:
+            brids.update([br.id for br in b.requests])
+        self.db.claim_buildrequests(now, self.master_name,
+                                    self.master_incarnation, brids)
+
     def getBuild(self, number):
         for b in self.building:
-            if b.build_status.number == number:
+            if b.build_status and b.build_status.number == number:
                 return b
         for b in self.old_building.keys():
-            if b.build_status.number == number:
+            if b.build_status and b.build_status.number == number:
                 return b
         return None
 
@@ -542,7 +686,7 @@ class Builder(pb.Referenceable):
         watchers = self.watchers[name]
         self.watchers[name] = []
         for w in watchers:
-            reactor.callLater(0, w.callback, fire_with)
+            eventually(w.callback, fire_with)
 
     def addLatentSlave(self, slave):
         assert interfaces.ILatentBuildSlave.providedBy(slave)
@@ -554,7 +698,7 @@ class Builder(pb.Referenceable):
             self.builder_status.addPointEvent(
                 ['added', 'latent', slave.slavename])
             self.slaves.append(sb)
-            reactor.callLater(0, self.maybeStartBuild)
+            self.triggerNewBuildCheck()
 
     def attached(self, slave, remote, commands):
         """This is invoked by the BuildSlave when the self.slavename bot
@@ -604,7 +748,6 @@ class Builder(pb.Referenceable):
         self.builder_status.addPointEvent(['connect', sb.slave.slavename])
         self.attaching_slaves.remove(sb)
         self.slaves.append(sb)
-        reactor.callLater(0, self.maybeStartBuild)
 
         self.fireTestEvent('attach')
         return self
@@ -614,14 +757,14 @@ class Builder(pb.Referenceable):
         # TODO: make this .addSlaveEvent?
         # TODO: remove from self.slaves (except that detached() should get
         #       run first, right?)
+        print why
         self.builder_status.addPointEvent(['failed', 'connect',
-                                           slave.slave.slavename])
+                                           slave.slavename])
         # TODO: add an HTMLLogFile of the exception
         self.fireTestEvent('attach', why)
 
     def detached(self, slave):
         """This is called when the connection to the bot is lost."""
-        log.msg("%s.detached" % self, slave.slavename)
         for sb in self.attaching_slaves + self.slaves:
             if sb.slave == slave:
                 break
@@ -663,51 +806,6 @@ class Builder(pb.Referenceable):
             self.builder_status.setBigState("idle")
             self.fireTestEvent('idle')
 
-    def maybeStartBuild(self):
-        log.msg("maybeStartBuild %s: %s %s" %
-                (self, self.buildable, self.slaves))
-        if not self.buildable:
-            self.updateBigStatus()
-            return # nothing to do
-
-        # pick an idle slave
-        available_slaves = [sb for sb in self.slaves if sb.isAvailable()]
-        if not available_slaves:
-            log.msg("%s: want to start build, but we don't have a remote"
-                    % self)
-            self.updateBigStatus()
-            return
-        if self.CHOOSE_SLAVES_RANDOMLY:
-            # TODO prefer idle over latent? maybe other sorting preferences?
-            sb = random.choice(available_slaves)
-        else:
-            sb = available_slaves[0]
-
-        # there is something to build, and there is a slave on which to build
-        # it. Grab the oldest request, see if we can merge it with anything
-        # else.
-        req = self.buildable.pop(0)
-        self.builder_status.removeBuildRequest(req.status)
-        mergers = []
-        botmaster = self.botmaster
-        for br in self.buildable[:]:
-            if botmaster.shouldMergeRequests(self, req, br):
-                self.buildable.remove(br)
-                self.builder_status.removeBuildRequest(br.status)
-                mergers.append(br)
-        requests = [req] + mergers
-
-        # Create a new build from our build factory and set ourself as the
-        # builder.
-        build = self.buildFactory.newBuild(requests)
-        build.setBuilder(self)
-        build.setLocks(self.locks)
-        if len(self.env) > 0:
-            build.setSlaveEnvironment(self.env)
-
-        # start it
-        self.startBuild(build, sb)
-
     def startBuild(self, build, sb):
         """Start a build on the given slave.
         @param build: the L{base.Build} to start
@@ -720,28 +818,50 @@ class Builder(pb.Referenceable):
 
         self.building.append(build)
         self.updateBigStatus()
-        if isinstance(sb, LatentSlaveBuilder):
-            log.msg("starting build %s.. substantiating the slave %s" %
-                    (build, sb))
-            d = sb.substantiate(build)
-            def substantiated(res):
-                return sb.ping(self.START_BUILD_TIMEOUT)
-            def substantiation_failed(res):
-                self.builder_status.addPointEvent(
-                    ['removing', 'latent', sb.slave.slavename])
-                sb.slave.disconnect()
-                # TODO: should failover to a new Build
-                #self.retryBuild(sb.build)
-            d.addCallbacks(substantiated, substantiation_failed)
-        else:
-            log.msg("starting build %s.. pinging the slave %s" % (build, sb))
-            d = sb.ping(self.START_BUILD_TIMEOUT)
-        # ping the slave to make sure they're still there. If they're fallen
-        # off the map (due to a NAT timeout or something), this will fail in
-        # a couple of minutes, depending upon the TCP timeout. TODO: consider
-        # making this time out faster, or at least characterize the likely
-        # duration.
-        d.addCallback(self._startBuild_1, build, sb)
+        log.msg("starting build %s using slave %s" % (build, sb))
+        d = sb.prepare(self.builder_status)
+
+        def _prepared(ready):
+            # If prepare returns True then it is ready and we start a build
+            # If it returns false then we don't start a new build.
+            d = defer.succeed(ready)
+
+            if not ready:
+                #FIXME: We should perhaps trigger a check to see if there is
+                # any other way to schedule the work
+                log.msg("slave %s can't build %s after all" % (build, sb))
+
+                # release the slave. This will queue a call to maybeStartBuild, which
+                # will fire after other notifyOnDisconnect handlers have marked the
+                # slave as disconnected (so we don't try to use it again).
+                # sb.buildFinished()
+
+                log.msg("re-queueing the BuildRequest %s" % build)
+                self.building.remove(build)
+                self._resubmit_buildreqs(build).addErrback(log.err)
+
+                sb.slave.releaseLocks()
+                self.triggerNewBuildCheck()
+
+                return d
+
+            def _ping(ign):
+                # ping the slave to make sure they're still there. If they've
+                # fallen off the map (due to a NAT timeout or something), this
+                # will fail in a couple of minutes, depending upon the TCP
+                # timeout.
+                #
+                # TODO: This can unnecessarily suspend the starting of a build, in
+                # situations where the slave is live but is pushing lots of data to
+                # us in a build.
+                log.msg("starting build %s.. pinging the slave %s" % (build, sb))
+                return sb.ping()
+            d.addCallback(_ping)
+            d.addCallback(self._startBuild_1, build, sb)
+
+            return d
+
+        d.addCallback(_prepared)
         return d
 
     def _startBuild_1(self, res, build, sb):
@@ -764,12 +884,11 @@ class Builder(pb.Referenceable):
         # BuildStatus that it has started, which will announce it to the
         # world (through our BuilderStatus object, which is its parent).
         # Finally it will start the actual build process.
+        bids = [self.db.build_started(req.id, bs.number) for req in build.requests]
         d = build.startBuild(bs, self.expectations, sb)
-        d.addCallback(self.buildFinished, sb)
-        d.addErrback(log.err) # this shouldn't happen. if it does, the slave
-                              # will be wedged
-        for req in build.requests:
-            req.buildStarted(build, bs)
+        d.addCallback(self.buildFinished, sb, bids)
+        # this shouldn't happen. if it does, the slave will be wedged
+        d.addErrback(log.err)
         return build # this is the IBuildControl
 
     def _startBuildFailed(self, why, build, sb):
@@ -783,13 +902,15 @@ class Builder(pb.Referenceable):
 
         log.msg("re-queueing the BuildRequest")
         self.building.remove(build)
-        for req in build.requests:
-            self.buildable.insert(0, req) # the interrupted build gets first
-                                          # priority
-            self.builder_status.addBuildRequest(req.status)
+        self._resubmit_buildreqs(build).addErrback(log.err)
 
+    def setupProperties(self, props):
+        props.setProperty("buildername", self.name, "Builder")
+        if len(self.properties) > 0:
+            for propertyname in self.properties:
+                props.setProperty(propertyname, self.properties[propertyname], "Builder")
 
-    def buildFinished(self, build, sb):
+    def buildFinished(self, build, sb, bids):
         """This is called when the Build has finished (either success or
         failure). Any exceptions during the build are reported with
         results=FAILURE, not with an errback."""
@@ -797,9 +918,24 @@ class Builder(pb.Referenceable):
         # by the time we get here, the Build has already released the slave
         # (which queues a call to maybeStartBuild)
 
+        self.db.builds_finished(bids)
+
+        results = build.build_status.getResults()
         self.building.remove(build)
-        for req in build.requests:
-            req.finished(build.build_status)
+        if results == RETRY:
+            self._resubmit_buildreqs(build).addErrback(log.err) # returns Deferred
+        else:
+            brids = [br.id for br in build.requests]
+            self.db.retire_buildrequests(brids, results)
+
+        if sb.slave:
+            sb.slave.releaseLocks()
+
+        self.triggerNewBuildCheck()
+
+    def _resubmit_buildreqs(self, build):
+        brids = [br.id for br in build.requests]
+        return self.db.resubmit_buildrequests(brids)
 
     def setExpectations(self, progress):
         """Mark the build as successful and update expectations for the next
@@ -822,45 +958,50 @@ class Builder(pb.Referenceable):
             self.remote.callRemote("shutdown")
 
 
-class BuilderControl(components.Adapter):
+class BuilderControl:
     implements(interfaces.IBuilderControl)
 
-    def requestBuild(self, req):
-        """Submit a BuildRequest to this Builder."""
-        self.original.submitBuildRequest(req)
+    def __init__(self, builder, parent):
+        self.original = builder
+        self.parent = parent # the IControl object
 
-    def requestBuildSoon(self, req):
-        """Submit a BuildRequest like requestBuild, but raise a
-        L{buildbot.interfaces.NoSlaveError} if no slaves are currently
-        available, so it cannot be used to queue a BuildRequest in the hopes
-        that a slave will eventually connect. This method is appropriate for
-        use by things like the web-page 'Force Build' button."""
-        if not self.original.slaves:
-            raise interfaces.NoSlaveError
-        self.requestBuild(req)
+    def submitBuildRequest(self, ss, reason, props=None, now=False):
+        bss = self.parent.submitBuildSet([self.original.name], ss, reason,
+                                         props, now)
+        brs = bss.getBuildRequests()[0]
+        return brs
 
-    def resubmitBuild(self, bs, reason="<rebuild, no reason given>"):
+    def rebuildBuild(self, bs, reason="<rebuild, no reason given>", extraProperties=None):
         if not bs.isFinished():
             return
 
         ss = bs.getSourceStamp(absolute=True)
-        req = base.BuildRequest(reason, ss, self.original.name)
-        self.requestBuild(req)
+        # Make a copy so as not to modify the original build.
+        properties = Properties()
+        # Don't include runtime-set properties in a rebuild request
+        properties.updateFromPropertiesNoRuntime(bs.getProperties())
+        if extraProperties is None:
+            properties.updateFromProperties(extraProperties)
+        self.submitBuildRequest(ss, reason, props=properties)
 
     def getPendingBuilds(self):
         # return IBuildRequestControl objects
-        raise NotImplementedError
+        retval = []
+        for r in self.original.getBuildable():
+            retval.append(BuildRequestControl(self.original, r))
+
+        return retval
 
     def getBuild(self, number):
         return self.original.getBuild(number)
 
-    def ping(self, timeout=30):
+    def ping(self):
         if not self.original.slaves:
             self.original.builder_status.addPointEvent(["ping", "no slave"])
             return defer.succeed(False) # interfaces.NoSlaveError
         dl = []
         for s in self.original.slaves:
-            dl.append(s.ping(timeout, self.original.builder_status))
+            dl.append(s.ping(self.original.builder_status))
         d = defer.DeferredList(dl)
         d.addCallback(self._gatherPingResults)
         return d
@@ -871,4 +1012,19 @@ class BuilderControl(components.Adapter):
                 return False
         return True
 
-components.registerAdapter(BuilderControl, Builder, interfaces.IBuilderControl)
+class BuildRequestControl:
+    implements(interfaces.IBuildRequestControl)
+
+    def __init__(self, builder, request):
+        self.original_builder = builder
+        self.original_request = request
+        self.brid = request.id
+
+    def subscribe(self, observer):
+        raise NotImplementedError
+
+    def unsubscribe(self, observer):
+        raise NotImplementedError
+
+    def cancel(self):
+        self.original_builder.cancelBuildRequest(self.brid)
