@@ -1,5 +1,7 @@
 # -*- test-case-name: buildbot.test.test_steps -*-
 
+import re
+
 from zope.interface import implements
 from twisted.internet import reactor, defer, error
 from twisted.protocols import basic
@@ -11,7 +13,7 @@ from twisted.web.util import formatFailure
 from buildbot import interfaces, locks
 from buildbot.status import progress
 from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, SKIPPED, \
-     EXCEPTION
+     EXCEPTION, RETRY, worst_status
 
 """
 BuildStep and RemoteCommand classes for master-side representation of the
@@ -53,8 +55,7 @@ class RemoteCommand(pb.Referenceable):
                                passed to
                                L{buildbot.slave.bot.SlaveBuilder.remote_startCommand}
                                and needs to have been registered
-                               slave-side by
-                               L{buildbot.slave.registry.registerSlaveCommand}
+                               slave-side in L{buildbot.slave.registry}
         @type  args:           dict
         @param args:           arguments to send to the remote command
         """
@@ -62,14 +63,6 @@ class RemoteCommand(pb.Referenceable):
         self.remote_command = remote_command
         self.args = args
 
-    def __getstate__(self):
-        dict = self.__dict__.copy()
-        # Remove the remote ref: if necessary (only for resumed builds), it
-        # will be reattached at resume time
-        if dict.has_key("remote"):
-            del dict["remote"]
-        return dict
-
     def run(self, step, remote):
         self.active = True
         self.step = step
@@ -103,12 +96,20 @@ class RemoteCommand(pb.Referenceable):
         @returns: a deferred that will fire when the remote command is
                   done (with None as the result)
         """
+
+        # Allow use of WithProperties in logfile path names.
+        cmd_args = self.args
+        if cmd_args.has_key("logfiles") and cmd_args["logfiles"]:
+            cmd_args = cmd_args.copy()
+            properties = self.step.build.getProperties()
+            cmd_args["logfiles"] = properties.render(cmd_args["logfiles"])
+
         # This method only initiates the remote command.
         # We will receive remote_update messages as the command runs.
         # We will get a single remote_complete when it finishes.
         # We should fire self.deferred when the command is done.
         d = self.remote.callRemote("startCommand", self, self.commandID,
-                                   self.remote_command, self.args)
+                                   self.remote_command, cmd_args)
         return d
 
     def interrupt(self, why):
@@ -256,6 +257,7 @@ class LoggedRemoteCommand(RemoteCommand):
 
     def __init__(self, *args, **kwargs):
         self.logs = {}
+        self.delayedLogs = {}
         self._closeWhenFinished = {}
         RemoteCommand.__init__(self, *args, **kwargs)
 
@@ -290,9 +292,15 @@ class LoggedRemoteCommand(RemoteCommand):
         if not logfileName:
             logfileName = loog.getName()
         assert logfileName not in self.logs
+        assert logfileName not in self.delayedLogs
         self.logs[logfileName] = loog
         self._closeWhenFinished[logfileName] = closeWhenFinished
 
+    def useLogDelayed(self, logfileName, activateCallBack, closeWhenFinished=False):
+        assert logfileName not in self.logs
+        assert logfileName not in self.delayedLogs
+        self.delayedLogs[logfileName] = (activateCallBack, closeWhenFinished)
+
     def start(self):
         log.msg("LoggedRemoteCommand.start")
         if 'stdio' not in self.logs:
@@ -313,6 +321,14 @@ class LoggedRemoteCommand(RemoteCommand):
             self.logs['stdio'].addHeader(data)
 
     def addToLog(self, logname, data):
+        # Activate delayed logs on first data.
+        if logname in self.delayedLogs:
+            (activateCallBack, closeWhenFinished) = self.delayedLogs[logname]
+            del self.delayedLogs[logname]
+            loog = activateCallBack(self)
+            self.logs[logname] = loog
+            self._closeWhenFinished[logname] = closeWhenFinished
+
         if logname in self.logs:
             self.logs[logname].addStdout(data)
         else:
@@ -431,9 +447,10 @@ class RemoteShellCommand(LoggedRemoteCommand):
     command is finished, it will fire a Deferred. You can then check the
     results of the command and parse the output however you like."""
 
-    def __init__(self, workdir, command, env=None, 
+    def __init__(self, workdir, command, env=None,
                  want_stdout=1, want_stderr=1,
-                 timeout=20*60, logfiles={}, usePTY="slave-config"):
+                 timeout=20*60, maxTime=None, logfiles={},
+                 usePTY="slave-config", logEnviron=True):
         """
         @type  workdir: string
         @param workdir: directory where the command ought to run,
@@ -472,6 +489,13 @@ class RemoteShellCommand(LoggedRemoteCommand):
                         produce any output for this number of seconds,
                         the command is hung and should be killed. Use
                         None to disable the timeout.
+
+        @param logEnviron: whether to log env vars on the slave side
+
+        @type  maxTime: int
+        @param maxTime: tell the remote that if the command fails to complete
+                        in this number of seconds, the command should be
+                        killed.  Use None to disable maxTime.
         """
 
         self.command = command # stash .command, set it later
@@ -486,7 +510,9 @@ class RemoteShellCommand(LoggedRemoteCommand):
                 'want_stderr': want_stderr,
                 'logfiles': logfiles,
                 'timeout': timeout,
+                'maxTime': maxTime,
                 'usePTY': usePTY,
+                'logEnviron': logEnviron,
                 }
         LoggedRemoteCommand.__init__(self, "shell", args)
 
@@ -545,7 +571,7 @@ class BuildStep:
     # immediately, the others will be taken into consideration when
     # determining the overall build status.
     #
-    # steps that are makred as alwaysRun will be run regardless of the outcome
+    # steps that are marked as alwaysRun will be run regardless of the outcome
     # of previous steps (especially steps with haltOnFailure=True)
     haltOnFailure = False
     flunkOnWarnings = False
@@ -594,6 +620,12 @@ class BuildStep:
             raise TypeError(why)
         self._pendingLogObservers = []
 
+        self._acquiringLock = None
+        self.stopped = False
+
+    def describe(self, done=False):
+        return [self.name]
+
     def setBuild(self, build):
         # subclasses which wish to base their behavior upon qualities of the
         # Build (e.g. use the list of changed files to run unit tests only on
@@ -639,8 +671,8 @@ class BuildStep:
     def getProperty(self, propname):
         return self.build.getProperty(propname)
 
-    def setProperty(self, propname, value, source="Step"):
-        self.build.setProperty(propname, value, source)
+    def setProperty(self, propname, value, source="Step", runtime=True):
+        self.build.setProperty(propname, value, source, runtime=runtime)
 
     def startStep(self, remote):
         """Begin the step. This returns a Deferred that will fire when the
@@ -686,29 +718,45 @@ class BuildStep:
                 log.msg("Hey, lock %s is claimed by both a Step (%s) and the"
                         " parent Build (%s)" % (l, self, self.build))
                 raise RuntimeError("lock claimed by both Step and Build")
+
+        # Set the step's text here so that the stepStarted notification sees
+        # the correct description
+        self.step_status.setText(self.describe(False))
+        self.step_status.stepStarted()
+
         d = self.acquireLocks()
         d.addCallback(self._startStep_2)
         return self.deferred
 
     def acquireLocks(self, res=None):
-        log.msg("acquireLocks(step %s, locks %s)" % (self, self.locks))
+        self._acquiringLock = None
         if not self.locks:
             return defer.succeed(None)
+        if self.stopped:
+            return defer.succeed(None)
+        log.msg("acquireLocks(step %s, locks %s)" % (self, self.locks))
         for lock, access in self.locks:
             if not lock.isAvailable(access):
+                self.step_status.setWaitingForLocks(True)
                 log.msg("step %s waiting for lock %s" % (self, lock))
                 d = lock.waitUntilMaybeAvailable(self, access)
                 d.addCallback(self.acquireLocks)
+                self._acquiringLock = (lock, access, d)
                 return d
         # all locks are available, claim them all
         for lock, access in self.locks:
             lock.claim(self, access)
+        self.step_status.setWaitingForLocks(False)
         return defer.succeed(None)
 
     def _startStep_2(self, res):
+        if self.stopped:
+            self.finished(FAILURE)
+            return
+
         if self.progress:
             self.progress.start()
-        self.step_status.stepStarted()
+
         try:
             skip = None
             if isinstance(self.doStepIf, bool):
@@ -721,6 +769,8 @@ class BuildStep:
                 skip = self.start()
 
             if skip == SKIPPED:
+                self.step_status.setText(self.describe(True) + ['skipped'])
+                self.step_status.setSkipped(True)
                 # this return value from self.start is a shortcut
                 # to finishing the step immediately
                 reactor.callLater(0, self.finished, SKIPPED)
@@ -796,12 +846,20 @@ class BuildStep:
         local processing should be skipped, and the Step completed with an
         error status. The results text should say something useful like
         ['step', 'interrupted'] or ['remote', 'lost']"""
-        pass
+        self.stopped = True
+        if self._acquiringLock:
+            lock, access, d = self._acquiringLock
+            lock.stopWaitingUntilAvailable(self, access, d)
+            d.callback(None)
 
     def releaseLocks(self):
         log.msg("releaseLocks(%s): %s" % (self, self.locks))
         for lock, access in self.locks:
-            lock.release(self, access)
+            if lock.isOwner(self, access):
+                lock.release(self, access)
+            else:
+                # This should only happen if we've been interrupted
+                assert self.stopped
 
     def finished(self, results):
         if self.progress:
@@ -869,7 +927,7 @@ class BuildStep:
         # This might change in the future (I might move away from CVS), but
         # if so I'll keep updating that string with suitably-comparable
         # values.
-        if sv.split(".") < minversion.split("."):
+        if map(int, sv.split(".")) < map(int, minversion.split(".")):
             return True
         return False
 
@@ -954,19 +1012,30 @@ class LoggingBuildStep(BuildStep):
     progressMetrics = ('output',)
     logfiles = {}
 
-    parms = BuildStep.parms + ['logfiles']
+    parms = BuildStep.parms + ['logfiles', 'lazylogfiles', 'log_eval_func']
+    cmd = None
 
-    def __init__(self, logfiles={}, *args, **kwargs):
+    def __init__(self, logfiles={}, lazylogfiles=False, log_eval_func=None,
+                 *args, **kwargs):
         BuildStep.__init__(self, *args, **kwargs)
-        self.addFactoryArguments(logfiles=logfiles)
+        self.addFactoryArguments(logfiles=logfiles,
+                                 lazylogfiles=lazylogfiles,
+                                 log_eval_func=log_eval_func)
         # merge a class-level 'logfiles' attribute with one passed in as an
         # argument
         self.logfiles = self.logfiles.copy()
         self.logfiles.update(logfiles)
+        self.lazylogfiles = lazylogfiles
+        assert not log_eval_func or callable(log_eval_func)
+        self.log_eval_func = log_eval_func
         self.addLogObserver('stdio', OutputProgressObserver("output"))
 
-    def describe(self, done=False):
-        raise NotImplementedError("implement this in a subclass")
+    def addLogFile(self, logname, filename):
+        """
+        This allows to add logfiles after construction, but before calling
+        startCommand().
+        """
+        self.logfiles[logname] = filename
 
     def startCommand(self, cmd, errorMessages=[]):
         """
@@ -1003,27 +1072,50 @@ class LoggingBuildStep(BuildStep):
 
     def setupLogfiles(self, cmd, logfiles):
         """Set up any additional logfiles= logs.
+
+        @param cmd: the LoggedRemoteCommand to add additional logs to.
+
+        @param logfiles: a dict of tuples (logname,remotefilename)
+                         specifying additional logs to watch. (note:
+                         the remotefilename component is currently
+                         ignored)
         """
         for logname,remotefilename in logfiles.items():
-            # tell the BuildStepStatus to add a LogFile
-            newlog = self.addLog(logname)
-            # and tell the LoggedRemoteCommand to feed it
-            cmd.useLog(newlog, True)
+            if self.lazylogfiles:
+                # Ask LoggedRemoteCommand to watch a logfile, but only add
+                # it when/if we see any data.
+                #
+                # The dummy default argument local_logname is a work-around for
+                # Python name binding; default values are bound by value, but
+                # captured variables in the body are bound by name.
+                callback = lambda cmd_arg, local_logname=logname: self.addLog(local_logname)
+                cmd.useLogDelayed(logname, callback, True)
+            else:
+                # tell the BuildStepStatus to add a LogFile
+                newlog = self.addLog(logname)
+                # and tell the LoggedRemoteCommand to feed it
+                cmd.useLog(newlog, True)
 
     def interrupt(self, reason):
         # TODO: consider adding an INTERRUPTED or STOPPED status to use
         # instead of FAILURE, might make the text a bit more clear.
         # 'reason' can be a Failure, or text
-        self.addCompleteLog('interrupt', str(reason))
-        d = self.cmd.interrupt(reason)
-        return d
+        BuildStep.interrupt(self, reason)
+        if self.step_status.isWaitingForLocks():
+            self.addCompleteLog('interrupt while waiting for locks', str(reason))
+        else:
+            self.addCompleteLog('interrupt', str(reason))
+
+        if self.cmd:
+            d = self.cmd.interrupt(reason)
+            return d
 
     def checkDisconnect(self, f):
         f.trap(error.ConnectionLost)
         self.step_status.setText(self.describe(True) +
                                  ["failed", "slave", "lost"])
         self.step_status.setText2(["failed", "slave", "lost"])
-        return self.finished(FAILURE)
+        return self.finished(RETRY)
 
     # to refine the status output, override one or more of the following
     # methods. Change as little as possible: start with the first ones on
@@ -1067,9 +1159,10 @@ class LoggingBuildStep(BuildStep):
         Override this to, say, declare WARNINGS if there is any stderr
         activity, or to say that rc!=0 is not actually an error."""
 
+        if self.log_eval_func:
+            return self.log_eval_func(cmd, self.step_status)
         if cmd.rc != 0:
             return FAILURE
-        # if cmd.log.getStderr(): return WARNINGS
         return SUCCESS
 
     def getText(self, cmd, results):
@@ -1110,7 +1203,32 @@ class LoggingBuildStep(BuildStep):
         self.step_status.setText(self.getText(cmd, results))
         self.step_status.setText2(self.maybeGetText2(cmd, results))
 
-# (WithProeprties used to be available in this module)
+
+# Parses the logs for a list of regexs. Meant to be invoked like:
+# regexes = ((re.compile(...), FAILURE), (re.compile(...), WARNINGS))
+# self.addStep(ShellCommand,
+#   command=...,
+#   ...,
+#   log_eval_func=lambda c,s: regex_log_evaluator(c, s, regexs)
+# )
+def regex_log_evaluator(cmd, step_status, regexes):
+    worst = SUCCESS
+    if cmd.rc != 0:
+        worst = FAILURE
+    for err, possible_status in regexes:
+        # worst_status returns the worse of the two status' passed to it.
+        # we won't be changing "worst" unless possible_status is worse than it,
+        # so we don't even need to check the log if that's the case
+        if worst_status(worst, possible_status) == possible_status:
+            if isinstance(err, (basestring)):
+                err = re.compile(".*%s.*" % err, re.DOTALL)
+            for l in cmd.logs.values():
+                if err.search(l.getText()):
+                    worst = possible_status
+    return worst
+
+
+# (WithProperties used to be available in this module)
 from buildbot.process.properties import WithProperties
 _hush_pyflakes = [WithProperties]
 del _hush_pyflakes
