@@ -3,6 +3,7 @@
 # protocols: irc, etc)
 
 import re, shlex
+from string import join, capitalize, lower
 
 from zope.interface import Interface, implements
 from twisted.internet import protocol, reactor
@@ -12,13 +13,18 @@ from twisted.application import internet
 
 from buildbot import interfaces, util
 from buildbot import version
+from buildbot.interfaces import IStatusReceiver
 from buildbot.sourcestamp import SourceStamp
-from buildbot.process.base import BuildRequest
 from buildbot.status import base
 from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, EXCEPTION
 from buildbot.scripts.runner import ForceOptions
 
-from string import join, capitalize, lower
+# twisted.internet.ssl requires PyOpenSSL, so be resilient if it's missing
+try:
+    from twisted.internet import ssl
+    have_ssl = True
+except ImportError:
+    have_ssl = False
 
 class UsageError(ValueError):
     def __init__(self, string = "Invalid usage", *more):
@@ -38,12 +44,11 @@ class IrcBuildRequest:
             self.parent.send("The build has been queued, I'll give a shout"
                              " when it starts")
 
-    def started(self, c):
+    def started(self, s):
         self.hasStarted = True
         if self.timer:
             self.timer.cancel()
             del self.timer
-        s = c.getStatus()
         eta = s.getETA()
         response = "build #%d forced" % s.getNumber()
         if eta is not None:
@@ -53,8 +58,8 @@ class IrcBuildRequest:
         d = s.waitUntilFinished()
         d.addCallback(self.parent.watchedBuildFinished)
 
-
-class Contact:
+class Contact(base.StatusReceiver):
+    implements(IStatusReceiver)
     """I hold the state for a single user's interaction with the buildbot.
 
     This base class provides all the basic behavior (the queries and
@@ -67,10 +72,12 @@ class Contact:
     """
 
     def __init__(self, channel):
+        #StatusReceiver.__init__(self) doesn't exist
         self.channel = channel
         self.notify_events = {}
         self.subscribed = 0
-	self.add_notification_events(channel.notify_events)
+        self.reported_builds = [] # tuples (when, buildername, buildnum)
+        self.add_notification_events(channel.notify_events)
 
     silly = {
         "What happen ?": "Somebody set up us the bomb.",
@@ -121,6 +128,22 @@ class Contact:
         minutes = minutes - 60*hours
         return "%dh%02dm%02ds" % (hours, minutes, seconds)
 
+    def reportBuild(self, builder, buildnum):
+        """Returns True if this build should be reported for this contact
+        (eliminating duplicates), and also records the report for later"""
+        for w, b, n in self.reported_builds:
+            if b == builder and n == buildnum:
+                return False
+        self.reported_builds.append([util.now(), builder, buildnum])
+
+        # clean the reported builds
+        horizon = util.now() - 60
+        while self.reported_builds and self.reported_builds[0][0] < horizon:
+            self.reported_builds.pop(0)
+
+        # and return True, since this is a new one
+        return True
+
     def doSilly(self, message):
         response = self.silly[message]
         if type(response) != type([]):
@@ -137,7 +160,7 @@ class Contact:
         self.send("buildbot-%s at your service" % version)
 
     def command_LIST(self, args, who):
-        args = args.split()
+        args = shlex.split(args)
         if len(args) == 0:
             raise UsageError, "try 'list builders'"
         if args[0] == 'builders':
@@ -155,7 +178,7 @@ class Contact:
     command_LIST.usage = "list builders - List configured builders"
 
     def command_STATUS(self, args, who):
-        args = args.split()
+        args = shlex.split(args)
         if len(args) == 0:
             which = "all"
         elif len(args) == 1:
@@ -214,7 +237,7 @@ class Contact:
             self.unsubscribe_from_build_events()
 
     def command_NOTIFY(self, args, who):
-        args = args.split()
+        args = shlex.split(args)
 
         if not args:
             raise UsageError("try 'notify on|off|list <EVENT>'")
@@ -245,7 +268,7 @@ class Contact:
     command_NOTIFY.usage = "notify on|off|list [<EVENT>] ... - Notify me about build events.  event should be one or more of: 'started', 'finished', 'failure', 'success', 'exception' or 'xToY' (where x and Y are one of success, warnings, failure, exception, but Y is capitalized)"
 
     def command_WATCH(self, args, who):
-        args = args.split()
+        args = shlex.split(args)
         if len(args) != 1:
             raise UsageError("try 'watch <builder>'")
         which = args[0]
@@ -278,8 +301,8 @@ class Contact:
         log.msg('[Contact] Builder %s changed state to %s' % (builderName, state))
 
     def requestSubmitted(self, brstatus):
-        log.msg('[Contact] BuildRequest for %s submiitted to Builder %s' % 
-            (brstatus.getSourceStamp(), brstatus.builderName))
+        log.msg('[Contact] BuildRequest %d for %s submitted' %
+            (brstatus.brid, brstatus.getSourceStamp()))
 
     def builderRemoved(self, builderName):
         log.msg('[Contact] Builder %s removed' % (builderName))
@@ -299,67 +322,76 @@ class Contact:
             log.msg('Not notifying for a build when started-notification disabled')
             return
 
-        r = "build #%d of %s started" % \
+        r = "build #%d of %s started, including [%s]" % \
            (build.getNumber(),
-             builder.getName())
-
-        r += " including [" + ", ".join(map(lambda c: repr(c.revision), build.getChanges())) + "]"
+            builder.getName(),
+            ", ".join([str(c.revision) for c in build.getChanges()])
+            )
 
         self.send(r)
 
+    results_descriptions = {
+        SUCCESS: "Success",
+        WARNINGS: "Warnings",
+        FAILURE: "Failure",
+        EXCEPTION: "Exception",
+        }
+
     def buildFinished(self, builderName, build, results):
         builder = build.getBuilder()
 
-        results_descriptions = {
-            SUCCESS: "Success",
-            WARNINGS: "Warnings",
-            FAILURE: "Failure",
-            EXCEPTION: "Exception",
-            }
-
         # only notify about builders we are interested in
         log.msg('[Contact] builder %r in category %s finished' % (builder, builder.category))
 
-        if self.notify_for('started'):
-            return
-
         if (self.channel.categories != None and
             builder.category not in self.channel.categories):
             return
 
+        if not self.notify_for_finished(build):
+            return
+
+        builder_name = builder.getName()
+        buildnum = build.getNumber()
+
+        if self.reportBuild(builder_name, buildnum):
+            r = "build #%d of %s is complete: %s" % \
+                (buildnum,
+                 builder_name,
+                 self.results_descriptions.get(build.getResults(), "??"))
+            r += " [%s]" % " ".join(build.getText())
+            buildurl = self.channel.status.getURLForThing(build)
+            if buildurl:
+                r += "  Build details are at %s" % buildurl
+
+            if self.channel.showBlameList and build.getResults() != SUCCESS and len(build.changes) != 0:
+                r += '  blamelist: ' + ', '.join(list(set([c.who for c in build.changes])))
+
+            self.send(r)
+
+    def notify_for_finished(self, build):
         results = build.getResults()
 
-        r = "build #%d of %s is complete: %s" % \
-            (build.getNumber(),
-             builder.getName(),
-             results_descriptions.get(results, "??"))
-        r += " [%s]" % " ".join(build.getText())
-        buildurl = self.channel.status.getURLForThing(build)
-        if buildurl:
-            r += "  Build details are at %s" % buildurl
+        if self.notify_for('finished'):
+            return True
 
-        if self.notify_for('finished') or self.notify_for(lower(results_descriptions.get(results))):
-            self.send(r)
-            return
+        if self.notify_for(lower(self.results_descriptions.get(results))):
+            return True
 
         prevBuild = build.getPreviousBuild()
         if prevBuild:
             prevResult = prevBuild.getResults()
 
-            required_notification_control_string = join((lower(results_descriptions.get(prevResult)), \
+            required_notification_control_string = join((lower(self.results_descriptions.get(prevResult)), \
                                                              'To', \
-                                                             capitalize(results_descriptions.get(results))), \
+                                                             capitalize(self.results_descriptions.get(results))), \
                                                             '')
 
             if (self.notify_for(required_notification_control_string)):
-                self.send(r)
+                return True
+
+        return False
 
     def watchedBuildFinished(self, b):
-        results = {SUCCESS: "Success",
-                   WARNINGS: "Warnings",
-                   FAILURE: "Failure",
-                   EXCEPTION: "Exception",
-                   }
 
         # only notify about builders we are interested in
         builder = b.getBuilder()
@@ -369,18 +401,22 @@ class Contact:
             builder.category not in self.channel.categories):
             return
 
-        r = "Hey! build %s #%d is complete: %s" % \
-            (b.getBuilder().getName(),
-             b.getNumber(),
-             results.get(b.getResults(), "??"))
-        r += " [%s]" % " ".join(b.getText())
-        self.send(r)
-        buildurl = self.channel.status.getURLForThing(b)
-        if buildurl:
-            self.send("Build details are at %s" % buildurl)
+        builder_name = b.getBuilder().getName()
+        buildnum = b.getNumber()
+
+        if self.reportBuild(builder_name, buildnum):
+            r = "Hey! build %s #%d is complete: %s" % \
+                (builder_name, 
+                 buildnum,
+                 self.results_descriptions.get(b.getResults(), "??"))
+            r += " [%s]" % " ".join(b.getText())
+            self.send(r)
+            buildurl = self.channel.status.getURLForThing(b)
+            if buildurl:
+                self.send("Build details are at %s" % buildurl)
 
     def command_FORCE(self, args, who):
-        args = shlex.split(args) # TODO: this requires python2.3 or newer
+        args = shlex.split(args)
         if not args:
             raise UsageError("try 'force build WHICH <REASON>'")
         what = args.pop(0)
@@ -388,7 +424,7 @@ class Contact:
             raise UsageError("try 'force build WHICH <REASON>'")
         opts = ForceOptions()
         opts.parseOptions(args)
-        
+
         which = opts['builder']
         branch = opts['branch']
         revision = opts['revision']
@@ -398,8 +434,8 @@ class Contact:
             raise UsageError("you must provide a Builder, "
                              "try 'force build WHICH <REASON>'")
 
-        # keep weird stuff out of the branch and revision strings. TODO:
-        # centralize this somewhere.
+        # keep weird stuff out of the branch and revision strings. 
+        # TODO:  centralize this somewhere.
         if branch and not re.match(r'^[\w\.\-\/]*$', branch):
             log.msg("bad branch '%s'" % branch)
             self.send("sorry, bad branch '%s'" % branch)
@@ -411,24 +447,23 @@ class Contact:
 
         bc = self.getControl(which)
 
-        r = "forced: by %s: %s" % (self.describeUser(who), reason)
         # TODO: maybe give certain users the ability to request builds of
         # certain branches
-        s = SourceStamp(branch=branch, revision=revision)
-        req = BuildRequest(r, s, which)
+        reason = "forced: by %s: %s" % (self.describeUser(who), reason)
+        ss = SourceStamp(branch=branch, revision=revision)
         try:
-            bc.requestBuildSoon(req)
+            brs = bc.submitBuildRequest(ss, reason, now=True)
         except interfaces.NoSlaveError:
             self.send("sorry, I can't force a build: all slaves are offline")
             return
         ireq = IrcBuildRequest(self)
-        req.subscribe(ireq.started)
+        brs.subscribe(ireq.started)
 
 
     command_FORCE.usage = "force build <which> <reason> - Force a build"
 
     def command_STOP(self, args, who):
-        args = args.split(None, 2)
+        args = shlex.split(args)
         if len(args) < 3 or args[0] != 'build':
             raise UsageError, "try 'stop build WHICH <REASON>'"
         which = args[1]
@@ -494,7 +529,7 @@ class Contact:
         self.send("last build [%s]: %s" % (which, str))
 
     def command_LAST(self, args, who):
-        args = args.split()
+        args = shlex.split(args)
         if len(args) == 0:
             which = "all"
         elif len(args) == 1:
@@ -518,7 +553,7 @@ class Contact:
         return commands
 
     def command_HELP(self, args, who):
-        args = args.split()
+        args = shlex.split(args)
         if len(args) == 0:
             self.send("Get help on what? (try 'help <foo>', or 'commands' for a command list)")
             return
@@ -573,6 +608,8 @@ class Contact:
         reactor.callLater(timeout, self.act, response)
 
 class IRCContact(Contact):
+    implements(IStatusReceiver)
+
     # this is the IRC-specific subclass of Contact
 
     def __init__(self, channel, dest):
@@ -584,32 +621,18 @@ class IRCContact(Contact):
         self.dest = dest
 
     def describeUser(self, user):
-        if self.dest[0] == "#":
+        if self.dest[0] == '#':
             return "IRC user <%s> on channel %s" % (user, self.dest)
         return "IRC user <%s> (privmsg)" % user
 
     # userJoined(self, user, channel)
 
     def send(self, message):
-        self.channel.msg(self.dest, message.encode("ascii", "replace"))
+        self.channel.msgOrNotice(self.dest, message.encode("ascii", "replace"))
+
     def act(self, action):
         self.channel.me(self.dest, action.encode("ascii", "replace"))
 
-    def command_JOIN(self, args, who):
-        args = args.split()
-        to_join = args[0]
-        self.channel.join(to_join)
-        self.send("Joined %s" % to_join)
-    command_JOIN.usage = "join channel - Join another channel"
-
-    def command_LEAVE(self, args, who):
-        args = args.split()
-        to_leave = args[0]
-        self.send("Buildbot has been told to leave %s" % to_leave)
-        self.channel.part(to_leave)
-    command_LEAVE.usage = "leave channel - Leave a channel"
-
-
     def handleMessage(self, message, who):
         # a message has arrived from 'who'. For broadcast contacts (i.e. when
         # people do an irc 'buildbot: command'), this will be a string
@@ -664,8 +687,9 @@ class IrcStatusBot(irc.IRCClient):
     """I represent the buildbot to an IRC server.
     """
     implements(IChannel)
+    contactClass = IRCContact
 
-    def __init__(self, nickname, password, channels, status, categories, notify_events):
+    def __init__(self, nickname, password, channels, status, categories, notify_events, noticeOnChannel = False, showBlameList = False):
         """
         @type  nickname: string
         @param nickname: the nickname by which this bot should be known
@@ -686,6 +710,14 @@ class IrcStatusBot(irc.IRCClient):
         self.counter = 0
         self.hasQuit = 0
         self.contacts = {}
+        self.noticeOnChannel = noticeOnChannel
+        self.showBlameList = showBlameList
+
+    def msgOrNotice(self, dest, message):
+        if self.noticeOnChannel and dest[0] == '#':
+            self.notice(dest, message)
+        else:
+            self.msg(dest, message)
 
     def addContact(self, name, contact):
         self.contacts[name] = contact
@@ -693,7 +725,7 @@ class IrcStatusBot(irc.IRCClient):
     def getContact(self, name):
         if name in self.contacts:
             return self.contacts[name]
-        new_contact = IRCContact(self, name)
+        new_contact = self.contactClass(self, name)
         self.contacts[name] = new_contact
         return new_contact
 
@@ -745,6 +777,9 @@ class IrcStatusBot(irc.IRCClient):
 
     def joined(self, channel):
         self.log("I have joined %s" % (channel,))
+        # trigger contact contructor, which in turn subscribes to notify events
+        self.getContact(channel)
+
     def left(self, channel):
         self.log("I have left %s" % (channel,))
     def kickedFrom(self, channel, kicker, message):
@@ -777,14 +812,16 @@ class IrcStatusFactory(ThrottledClientFactory):
     shuttingDown = False
     p = None
 
-    def __init__(self, nickname, password, channels, categories, notify_events):
+    def __init__(self, nickname, password, channels, categories, notify_events, noticeOnChannel = False, showBlameList = False):
         #ThrottledClientFactory.__init__(self) # doesn't exist
         self.status = None
         self.nickname = nickname
         self.password = password
         self.channels = channels
         self.categories = categories
-	self.notify_events = notify_events
+        self.notify_events = notify_events
+        self.noticeOnChannel = noticeOnChannel
+        self.showBlameList = showBlameList
 
     def __getstate__(self):
         d = self.__dict__.copy()
@@ -799,7 +836,9 @@ class IrcStatusFactory(ThrottledClientFactory):
     def buildProtocol(self, address):
         p = self.protocol(self.nickname, self.password,
                           self.channels, self.status,
-                          self.categories, self.notify_events)
+                          self.categories, self.notify_events,
+                          noticeOnChannel = self.noticeOnChannel,
+                          showBlameList = self.showBlameList)
         p.factory = self
         p.status = self.status
         p.control = self.control
@@ -823,16 +862,21 @@ class IrcStatusFactory(ThrottledClientFactory):
 
 
 class IRC(base.StatusReceiverMultiService):
+    implements(IStatusReceiver)
     """I am an IRC bot which can be queried for status information. I
     connect to a single IRC server and am known by a single nickname on that
     server, however I can join multiple channels."""
 
+    in_test_harness = False
+
     compare_attrs = ["host", "port", "nick", "password",
-                     "channels", "allowForce",
+                     "channels", "allowForce", "useSSL",
                      "categories"]
 
     def __init__(self, host, nick, channels, port=6667, allowForce=True,
-                 categories=None, password=None, notify_events={}):
+                 categories=None, password=None, notify_events={},
+                 noticeOnChannel = False, showBlameList = True,
+                 useSSL=False):
         base.StatusReceiverMultiService.__init__(self)
 
         assert allowForce in (True, False) # TODO: implement others
@@ -845,13 +889,26 @@ class IRC(base.StatusReceiverMultiService):
         self.password = password
         self.allowForce = allowForce
         self.categories = categories
-	self.notify_events = notify_events
-
-        # need to stash the factory so we can give it the status object
+        self.notify_events = notify_events
+        log.msg('Notify events %s' % notify_events)
         self.f = IrcStatusFactory(self.nick, self.password,
-                                  self.channels, self.categories, self.notify_events)
+                                  self.channels, self.categories, self.notify_events,
+                                  noticeOnChannel = noticeOnChannel,
+                                  showBlameList = showBlameList)
+
+        # don't set up an actual ClientContextFactory if we're running tests.
+        if self.in_test_harness:
+            return
+
+        if useSSL:
+            # SSL client needs a ClientContextFactory for some SSL mumbo-jumbo
+            if not have_ssl:
+                raise RuntimeError("useSSL requires PyOpenSSL")
+            cf = ssl.ClientContextFactory()
+            c = internet.SSLClient(self.host, self.port, self.f, cf)
+        else:
+            c = internet.TCPClient(self.host, self.port, self.f)
 
-        c = internet.TCPClient(host, port, self.f)
         c.setServiceParent(self)
 
     def setServiceParent(self, parent):
@@ -873,3 +930,4 @@ class IRC(base.StatusReceiverMultiService):
 ## buildbot: status full-2.3
 ##  building, not, % complete, ETA
 ## buildbot: force build full-2.3 "reason"
+
