@@ -2,8 +2,11 @@
 
 import re
 from twisted.python import log
+from twisted.spread import pb
 from buildbot.process.buildstep import LoggingBuildStep, RemoteShellCommand
+from buildbot.process.buildstep import RemoteCommand
 from buildbot.status.builder import SUCCESS, WARNINGS, FAILURE, STDOUT, STDERR
+from buildbot.interfaces import BuildSlaveTooOldError
 
 # for existing configurations that import WithProperties from here.  We like
 # to move this class around just to keep our readers guessing.
@@ -39,6 +42,12 @@ class ShellCommand(LoggingBuildStep):
                     something approximating real-time. (note that logfiles=
                     is actually handled by our parent class LoggingBuildStep)
 
+    @ivar lazylogfiles: Defaults to False. If True, logfiles will be tracked
+                        `lazily', meaning they will only be added when and if
+                        they are written to. Empty or nonexistent logfiles
+                        will be omitted. (Also handled by class
+                        LoggingBuildStep.)
+
     """
 
     name = "shell"
@@ -93,6 +102,9 @@ class ShellCommand(LoggingBuildStep):
         # we need to stash the RemoteShellCommand's args too
         self.addFactoryArguments(**kwargs)
 
+    def setStepStatus(self, step_status):
+        LoggingBuildStep.setStepStatus(self, step_status)
+
     def setDefaultWorkdir(self, workdir):
         rkw = self.remote_kwargs
         rkw['workdir'] = rkw['workdir'] or workdir
@@ -117,24 +129,30 @@ class ShellCommand(LoggingBuildStep):
                      'tests' ...)
         """
 
-        if done and self.descriptionDone is not None:
-            return list(self.descriptionDone)
-        if self.description is not None:
-            return list(self.description)
+        try:
+            properties = self.build.getProperties()
 
-        properties = self.build.getProperties()
-        words = self.command
-        if isinstance(words, (str, unicode)):
-            words = words.split()
-        # render() each word to handle WithProperties objects
-        words = properties.render(words)
-        if len(words) < 1:
+            if done and self.descriptionDone is not None:
+                return properties.render(self.descriptionDone)
+            if self.description is not None:
+                return properties.render(self.description)
+
+            words = self.command
+            if isinstance(words, (str, unicode)):
+                words = words.split()
+            # render() each word to handle WithProperties objects
+            words = properties.render(words)
+            if len(words) < 1:
+                return ["???"]
+            if len(words) == 1:
+                return ["'%s'" % words[0]]
+            if len(words) == 2:
+                return ["'%s" % words[0], "%s'" % words[1]]
+            return ["'%s" % words[0], "%s" % words[1], "...'"]
+        except:
+            log.msg("Error describing step")
+            log.err()
             return ["???"]
-        if len(words) == 1:
-            return ["'%s'" % words[0]]
-        if len(words) == 2:
-            return ["'%s" % words[0], "%s'" % words[1]]
-        return ["'%s" % words[0], "%s" % words[1], "...'"]
 
     def setupEnvironment(self, cmd):
         # merge in anything from Build.slaveEnvironment
@@ -191,12 +209,12 @@ class ShellCommand(LoggingBuildStep):
 
         # create the actual RemoteShellCommand instance now
         kwargs = properties.render(self.remote_kwargs)
-        kwargs['command'] = properties.render(self.command)
+        command = properties.render(self.command)
+        kwargs['command'] = command
         kwargs['logfiles'] = self.logfiles
 
         # check for the usePTY flag
         if kwargs.has_key('usePTY') and kwargs['usePTY'] != 'slave-config':
-            slavever = self.slaveVersion("shell", "old")
             if self.slaveVersionIsOlderThan("svn", "2.7"):
                 warnings.append("NOTE: slave does not allow master to override usePTY\n")
 
@@ -211,8 +229,13 @@ class ShellCommand(LoggingBuildStep):
 class TreeSize(ShellCommand):
     name = "treesize"
     command = ["du", "-s", "-k", "."]
+    description = "measuring tree size"
+    descriptionDone = "tree size measured"
     kib = None
 
+    def __init__(self, *args, **kwargs):
+        ShellCommand.__init__(self, *args, **kwargs)
+
     def commandComplete(self, cmd):
         out = cmd.logs['stdio'].getText()
         m = re.search(r'^(\d+)', out)
@@ -235,20 +258,13 @@ class TreeSize(ShellCommand):
 class SetProperty(ShellCommand):
     name = "setproperty"
 
-    def __init__(self, **kwargs):
-        self.property = None
-        self.extract_fn = None
-        self.strip = True
+    def __init__(self, property=None, extract_fn=None, strip=True, **kwargs):
+        self.property = property
+        self.extract_fn = extract_fn
+        self.strip = strip
 
-        if kwargs.has_key('property'):
-            self.property = kwargs['property']
-            del kwargs['property']
-        if kwargs.has_key('extract_fn'):
-            self.extract_fn = kwargs['extract_fn']
-            del kwargs['extract_fn']
-        if kwargs.has_key('strip'):
-            self.strip = kwargs['strip']
-            del kwargs['strip']
+        assert (property is not None) ^ (extract_fn is not None), \
+                "Exactly one of property and extract_fn must be set"
 
         ShellCommand.__init__(self, **kwargs)
 
@@ -256,9 +272,6 @@ class SetProperty(ShellCommand):
         self.addFactoryArguments(extract_fn=self.extract_fn)
         self.addFactoryArguments(strip=self.strip)
 
-        assert self.property or self.extract_fn, \
-            "SetProperty step needs either property= or extract_fn="
-
         self.property_changes = {}
 
     def commandComplete(self, cmd):
@@ -296,23 +309,198 @@ class Configure(ShellCommand):
     descriptionDone = ["configure"]
     command = ["./configure"]
 
+class StringFileWriter(pb.Referenceable):
+    """
+    FileWriter class that just puts received data into a buffer.
+
+    Used to upload a file from slave for inline processing rather than
+    writing into a file on master.
+    """
+    def __init__(self):
+        self.buffer = ""
+
+    def remote_write(self, data):
+        self.buffer += data
+
+    def remote_close(self):
+        pass
+
+class SilentRemoteCommand(RemoteCommand):
+    """
+    Remote command subclass used to run an internal file upload command on the
+    slave. We do not need any progress updates from such command, so override
+    remoteUpdate() with an empty method.
+    """
+    def remoteUpdate(self, update):
+        pass
+
 class WarningCountingShellCommand(ShellCommand):
     warnCount = 0
     warningPattern = '.*warning[: ].*'
+    # The defaults work for GNU Make.
+    directoryEnterPattern = "make.*: Entering directory [\"`'](.*)['`\"]"
+    directoryLeavePattern = "make.*: Leaving directory"
+    suppressionFile = None
 
-    def __init__(self, warningPattern=None, **kwargs):
+    commentEmptyLineRe = re.compile(r"^\s*(\#.*)?$")
+    suppressionLineRe = re.compile(r"^\s*(.+?)\s*:\s*(.+?)\s*(?:[:]\s*([0-9]+)(?:-([0-9]+))?\s*)?$")
+
+    def __init__(self, workdir=None,
+                 warningPattern=None, warningExtractor=None,
+                 directoryEnterPattern=None, directoryLeavePattern=None,
+                 suppressionFile=None, **kwargs):
+        self.workdir = workdir
         # See if we've been given a regular expression to use to match
         # warnings. If not, use a default that assumes any line with "warning"
         # present is a warning. This may lead to false positives in some cases.
         if warningPattern:
             self.warningPattern = warningPattern
+        if directoryEnterPattern:
+            self.directoryEnterPattern = directoryEnterPattern
+        if directoryLeavePattern:
+            self.directoryLeavePattern = directoryLeavePattern
+        if suppressionFile:
+            self.suppressionFile = suppressionFile
+        if warningExtractor:
+            self.warningExtractor = warningExtractor
+        else:
+            self.warningExtractor = WarningCountingShellCommand.warnExtractWholeLine
 
         # And upcall to let the base class do its work
-        ShellCommand.__init__(self, **kwargs)
+        ShellCommand.__init__(self, workdir=workdir, **kwargs)
 
-        self.addFactoryArguments(warningPattern=warningPattern)
+        self.addFactoryArguments(warningPattern=warningPattern,
+                                 directoryEnterPattern=directoryEnterPattern,
+                                 directoryLeavePattern=directoryLeavePattern,
+                                 warningExtractor=warningExtractor,
+                                 suppressionFile=suppressionFile)
+        self.suppressions = []
+        self.directoryStack = []
+
+    def setDefaultWorkdir(self, workdir):
+        if self.workdir is None:
+            self.workdir = workdir
+        ShellCommand.setDefaultWorkdir(self, workdir)
+
+    def addSuppression(self, suppressionList):
+        """
+        This method can be used to add patters of warnings that should
+        not be counted.
+
+        It takes a single argument, a list of patterns.
+
+        Each pattern is a 4-tuple (FILE-RE, WARN-RE, START, END).
+
+        FILE-RE is a regular expression (string or compiled regexp), or None.
+        If None, the pattern matches all files, else only files matching the
+        regexp. If directoryEnterPattern is specified in the class constructor,
+        matching is against the full path name, eg. src/main.c.
+
+        WARN-RE is similarly a regular expression matched against the
+        text of the warning, or None to match all warnings.
+
+        START and END form an inclusive line number range to match against. If
+        START is None, there is no lower bound, similarly if END is none there
+        is no upper bound."""
+
+        for fileRe, warnRe, start, end in suppressionList:
+            if fileRe != None and isinstance(fileRe, str):
+                fileRe = re.compile(fileRe)
+            if warnRe != None and isinstance(warnRe, str):
+                warnRe = re.compile(warnRe)
+            self.suppressions.append((fileRe, warnRe, start, end))
+
+    def warnExtractWholeLine(self, line, match):
+        """
+        Extract warning text as the whole line.
+        No file names or line numbers."""
+        return (None, None, line)
+
+    def warnExtractFromRegexpGroups(self, line, match):
+        """
+        Extract file name, line number, and warning text as groups (1,2,3)
+        of warningPattern match."""
+        file = match.group(1)
+        lineNo = match.group(2)
+        if lineNo != None:
+            lineNo = int(lineNo)
+        text = match.group(3)
+        return (file, lineNo, text)
+
+    def maybeAddWarning(self, warnings, line, match):
+        if self.suppressions:
+            (file, lineNo, text) = self.warningExtractor(self, line, match)
+
+            if file != None and file != "" and self.directoryStack:
+                currentDirectory = self.directoryStack[-1]
+                if currentDirectory != None and currentDirectory != "":
+                    file = "%s/%s" % (currentDirectory, file)
+
+            # Skip adding the warning if any suppression matches.
+            for fileRe, warnRe, start, end in self.suppressions:
+                if ( (file == None or fileRe == None or fileRe.search(file)) and
+                     (warnRe == None or  warnRe.search(text)) and
+                     ((start == None and end == None) or
+                      (lineNo != None and start <= lineNo and end >= lineNo)) ):
+                    return
+
+        warnings.append(line)
+        self.warnCount += 1
+
+    def start(self):
+        if self.suppressionFile == None:
+            return ShellCommand.start(self)
+
+        version = self.slaveVersion("uploadFile")
+        if not version:
+            m = "Slave is too old, does not know about uploadFile"
+            raise BuildSlaveTooOldError(m)
+
+        self.myFileWriter = StringFileWriter()
+
+        properties = self.build.getProperties()
+
+        args = {
+            'slavesrc': properties.render(self.suppressionFile),
+            'workdir': self.workdir,
+            'writer': self.myFileWriter,
+            'maxsize': None,
+            'blocksize': 32*1024,
+            }
+        cmd = SilentRemoteCommand('uploadFile', args)
+        d = self.runCommand(cmd)
+        d.addCallback(self.uploadDone)
+        d.addErrback(self.failed)
+
+    def uploadDone(self, dummy):
+        lines = self.myFileWriter.buffer.split("\n")
+        del(self.myFileWriter)
+
+        list = []
+        for line in lines:
+            if self.commentEmptyLineRe.match(line):
+                continue
+            match = self.suppressionLineRe.match(line)
+            if (match):
+                file, test, start, end = match.groups()
+                if (end != None):
+                    end = int(end)
+                if (start != None):
+                    start = int(start)
+                    if end == None:
+                        end = start
+                list.append((file, test, start, end))
+
+        self.addSuppression(list)
+        return ShellCommand.start(self)
 
     def createSummary(self, log):
+        """
+        Match log lines against warningPattern.
+
+        Warnings are collected into another log for this step, and the
+        build-wide 'warnings-count' is updated."""
+
         self.warnCount = 0
 
         # Now compile a regular expression from whichever warning pattern we're
@@ -324,6 +512,14 @@ class WarningCountingShellCommand(ShellCommand):
         if isinstance(wre, str):
             wre = re.compile(wre)
 
+        directoryEnterRe = self.directoryEnterPattern
+        if directoryEnterRe != None and isinstance(directoryEnterRe, str):
+            directoryEnterRe = re.compile(directoryEnterRe)
+
+        directoryLeaveRe = self.directoryLeavePattern
+        if directoryLeaveRe != None and isinstance(directoryLeaveRe, str):
+            directoryLeaveRe = re.compile(directoryLeaveRe)
+
         # Check if each line in the output from this command matched our
         # warnings regular expressions. If did, bump the warnings count and
         # add the line to the collection of lines with warnings
@@ -331,14 +527,24 @@ class WarningCountingShellCommand(ShellCommand):
         # TODO: use log.readlines(), except we need to decide about stdout vs
         # stderr
         for line in log.getText().split("\n"):
-            if wre.match(line):
-                warnings.append(line)
-                self.warnCount += 1
+            if directoryEnterRe:
+                match = directoryEnterRe.search(line)
+                if match:
+                    self.directoryStack.append(match.group(1))
+                if (directoryLeaveRe and
+                    self.directoryStack and
+                    directoryLeaveRe.search(line)):
+                        self.directoryStack.pop()
+
+            match = wre.match(line)
+            if match:
+                self.maybeAddWarning(warnings, line, match)
 
         # If there were any warnings, make the log if lines with warnings
         # available
         if self.warnCount:
-            self.addCompleteLog("warnings", "\n".join(warnings) + "\n")
+            self.addCompleteLog("warnings (%d)" % self.warnCount,
+                    "\n".join(warnings) + "\n")
 
         warnings_stat = self.step_status.getStatistic('warnings', 0)
         self.step_status.setStatistic('warnings', warnings_stat + self.warnCount)
@@ -433,12 +639,13 @@ class PerlModuleTest(Test):
         total = 0
         passed = 0
         failed = 0
-        rc = cmd.rc
+        rc = SUCCESS
+        if cmd.rc > 0:
+            rc = FAILURE
 
         # New version of Test::Harness?
-        try:
+        if "Test Summary Report" in lines:
             test_summary_report_index = lines.index("Test Summary Report")
-
             del lines[0:test_summary_report_index + 2]
 
             re_test_result = re.compile("^Result: (PASS|FAIL)$|Tests: \d+ Failed: (\d+)\)|Files=\d+, Tests=(\d+)")
@@ -447,16 +654,15 @@ class PerlModuleTest(Test):
             test_result_lines = [mo.groups() for mo in mos if mo]
 
             for line in test_result_lines:
-                if line[0] == 'PASS':
-                    rc = SUCCESS
-                elif line[0] == 'FAIL':
+                if line[0] == 'FAIL':
                     rc = FAILURE
-                elif line[1]:
+
+                if line[1]:
                     failed += int(line[1])
-                elif line[2]:
+                if line[2]:
                     total = int(line[2])
 
-        except ValueError: # Nope, it's the old version
+        else: # Nope, it's the old version
             re_test_result = re.compile("^(All tests successful)|(\d+)/(\d+) subtests failed|Files=\d+, Tests=(\d+),")
 
             mos = map(lambda line: re_test_result.search(line), lines)
@@ -472,8 +678,6 @@ class PerlModuleTest(Test):
 
                     test_totals_line = test_result_lines[1]
                     total_str = test_totals_line[3]
-                    
-                    rc = SUCCESS
                 else:
                     failed_str = test_result_line[1]
                     failed = int(failed_str)
@@ -484,9 +688,24 @@ class PerlModuleTest(Test):
 
                 total = int(total_str)
 
+        warnings = 0
+        if self.warningPattern:
+            wre = self.warningPattern
+            if isinstance(wre, str):
+                wre = re.compile(wre)
+
+            warnings = len([l for l in lines if wre.search(l)])
+
+            # Because there are two paths that are used to determine
+            # the success/fail result, I have to modify it here if
+            # there were warnings.
+            if rc == SUCCESS and warnings:
+                rc = WARNINGS
+
         if total:
             passed = total - failed
 
-            self.setTestResults(total=total, failed=failed, passed=passed)
+            self.setTestResults(total=total, failed=failed, passed=passed,
+                                warnings=warnings)
 
         return rc
