@@ -122,11 +122,13 @@ class Source(LoggingBuildStep):
             assert isinstance(repeats, int)
             assert repeats > 0
         self.args = {'mode': mode,
-                     'workdir': workdir,
                      'timeout': timeout,
                      'retry': retry,
                      'patch': None, # set during .start
                      }
+        # This will get added to args later, after properties are rendered
+        self.workdir = workdir
+
         self.alwaysUseLatest = alwaysUseLatest
 
         # Compute defaults for descriptions:
@@ -142,8 +144,11 @@ class Source(LoggingBuildStep):
         self.description = description
         self.descriptionDone = descriptionDone
 
+    def setStepStatus(self, step_status):
+        LoggingBuildStep.setStepStatus(self, step_status)
+
     def setDefaultWorkdir(self, workdir):
-        self.args['workdir'] = self.args['workdir'] or workdir
+        self.workdir = self.workdir or workdir
 
     def describe(self, done=False):
         if done:
@@ -160,6 +165,34 @@ class Source(LoggingBuildStep):
         self.checkoutDelay value."""
         return None
 
+    def computeRepositoryURL(self, repository):
+        '''
+        Helper function that the repository URL based on the parameter the
+        source step took and the Change 'repository' property
+        '''
+
+        assert not repository or callable(repository) or isinstance(repository, dict) or \
+            isinstance(repository, str) or isinstance(repository, unicode)
+
+        s = self.build.getSourceStamp()
+        props = self.build.getProperties()
+
+        if not repository:
+            assert s.repository
+            return str(s.repository)
+        else:
+            if callable(repository):
+                return str(props.render(repository(s.repository)))
+            elif isinstance(repository, dict):
+                return str(props.render(repository.get(s.repository)))
+            else: # string or unicode
+                try:
+                    repourl = str(repository % s.repository)
+                except TypeError:
+                    # that's the backward compatibility case
+                    repourl = props.render(repository)
+                return repourl
+
     def start(self):
         if self.notReally:
             log.msg("faking %s checkout/update" % self.name)
@@ -169,6 +202,10 @@ class Source(LoggingBuildStep):
                                 % self.name)
             return SKIPPED
 
+        # Allow workdir to be WithProperties
+        properties = self.build.getProperties()
+        self.args['workdir'] = properties.render(self.workdir)
+
         # what source stamp would this build like to use?
         s = self.build.getSourceStamp()
         # if branch is None, then use the Step's "default" branch
@@ -177,6 +214,10 @@ class Source(LoggingBuildStep):
         revision = s.revision
         if not revision and not self.alwaysUseLatest:
             revision = self.computeSourceRevision(s.changes)
+            # the revision property is currently None, so set it to something
+            # more interesting
+            self.setProperty('revision', str(revision), "Source")
+
         # if patch is None, then do not patch the tree after checkout
 
         # 'patch' is None or a tuple of (patchlevel, diff, root)
@@ -197,6 +238,76 @@ class Source(LoggingBuildStep):
 
 
 
+class BK(Source):
+    """I perform BitKeeper checkout/update operations."""
+    
+    name = 'bk'
+    
+    def __init__(self, bkurl=None, baseURL=None,
+                 directory=None, extra_args=None, **kwargs):
+        """
+        @type  bkurl: string
+        @param bkurl: the URL which points to the BitKeeper server.
+                 
+        @type  baseURL: string
+        @param baseURL: if branches are enabled, this is the base URL to
+                        which a branch name will be appended. It should
+                        probably end in a slash. Use exactly one of
+                        C{bkurl} and C{baseURL}.
+        """
+                        
+        self.bkurl = bkurl
+        self.baseURL = baseURL
+        self.extra_args = extra_args
+        
+        Source.__init__(self, **kwargs)
+        self.addFactoryArguments(bkurl=bkurl,
+                                 baseURL=baseURL,
+                                 directory=directory,
+                                 extra_args=extra_args,
+                                 )
+
+        if bkurl and baseURL:
+            raise ValueError("you must use exactly one of bkurl and baseURL")
+
+        
+    def computeSourceRevision(self, changes):
+        return changes.revision
+                       
+                       
+    def startVC(self, branch, revision, patch):
+
+        warnings = []
+        slavever = self.slaveVersion("bk")
+        if not slavever:
+            m = "slave does not have the 'bk' command"
+            raise BuildSlaveTooOldError(m)
+
+        if self.bkurl:
+            assert not branch # we need baseURL= to use branches
+            self.args['bkurl'] = self.computeRepositoryURL(self.bkurl)
+        else:
+            self.args['bkurl'] = self.computeRepositoryURL(self.baseURL) + branch
+        self.args['revision'] = revision
+        self.args['patch'] = patch
+        self.args['branch'] = branch
+        if self.extra_args is not None:
+            self.args['extra_args'] = self.extra_args
+
+        revstuff = []
+        revstuff.append("[branch]")
+        if revision is not None:
+            revstuff.append("r%s" % revision)
+        if patch is not None:
+            revstuff.append("[patch]")
+        self.description.extend(revstuff)
+        self.descriptionDone.extend(revstuff)
+
+        cmd = LoggedRemoteCommand("bk", self.args)
+        self.startCommand(cmd, warnings)
+
+
+
 class CVS(Source):
     """I do CVS checkout/update operations.
 
@@ -218,9 +329,9 @@ class CVS(Source):
     # parsing each line. Might be handy to have a hook in LogFile that gets
     # called with each complete line.
 
-    def __init__(self, cvsroot, cvsmodule,
+    def __init__(self, cvsroot=None, cvsmodule="",
                  global_options=[], branch=None, checkoutDelay=None,
-                 checkout_options=[],
+                 checkout_options=[], export_options=[], extra_options=[],
                  login=None,
                  **kwargs):
 
@@ -279,25 +390,39 @@ class CVS(Source):
         @param checkout_options: these arguments are inserted in the cvs
                                command line, after 'checkout' but before
                                branch or revision specifiers.
+
+        @type  export_options: list of strings
+        @param export_options: these arguments are inserted in the cvs
+                               command line, after 'export' but before
+                               branch or revision specifiers.
+
+        @type  extra_options: list of strings
+        @param extra_options: these arguments are inserted in the cvs
+                               command line, after 'checkout' or 'export' but before
+                               branch or revision specifiers.
                                """
 
         self.checkoutDelay = checkoutDelay
         self.branch = branch
+        self.cvsroot = cvsroot
 
         Source.__init__(self, **kwargs)
         self.addFactoryArguments(cvsroot=cvsroot,
                                  cvsmodule=cvsmodule,
                                  global_options=global_options,
                                  checkout_options=checkout_options,
+                                 export_options=export_options,
+                                 extra_options=extra_options,
                                  branch=branch,
                                  checkoutDelay=checkoutDelay,
                                  login=login,
                                  )
 
-        self.args.update({'cvsroot': cvsroot,
-                          'cvsmodule': cvsmodule,
+        self.args.update({'cvsmodule': cvsmodule,
                           'global_options': global_options,
                           'checkout_options':checkout_options,
+                          'export_options':export_options,
+                          'extra_options':extra_options,
                           'login': login,
                           })
 
@@ -330,8 +455,21 @@ class CVS(Source):
                 log.msg(m)
                 raise BuildSlaveTooOldError(m)
 
+        if self.slaveVersionIsOlderThan("cvs", "2.10"):
+            if self.args['extra_options'] or self.args['export_options']:
+                m = ("This buildslave (%s) does not support export_options "
+                     "or extra_options arguments to the CVS step."
+                     % (self.build.slavename))
+                log.msg(m)
+                raise BuildSlaveTooOldError(m)
+            # the unwanted args are empty, and will probably be ignored by
+            # the slave, but delete them just to be safe
+            del self.args['export_options']
+            del self.args['extra_options']
+
         if branch is None:
             branch = "HEAD"
+        self.args['cvsroot'] = self.computeRepositoryURL(self.cvsroot)
         self.args['branch'] = branch
         self.args['revision'] = revision
         self.args['patch'] = patch
@@ -364,6 +502,7 @@ class SVN(Source):
     """I perform Subversion checkout/update operations."""
 
     name = 'svn'
+    branch_placeholder = '%%BRANCH%%'
 
     def __init__(self, svnurl=None, baseURL=None, defaultBranch=None,
                  directory=None, username=None, password=None,
@@ -375,9 +514,7 @@ class SVN(Source):
                        combining the access method (HTTP, ssh, local file),
                        the repository host/port, the repository path, the
                        sub-tree within the repository, and the branch to
-                       check out. Using C{svnurl} does not enable builds of
-                       alternate branches: use C{baseURL} to enable this.
-                       Use exactly one of C{svnurl} and C{baseURL}.
+                       check out. Use exactly one of C{svnurl} and C{baseURL}.
 
         @param baseURL: if branches are enabled, this is the base URL to
                         which a branch name will be appended. It should
@@ -390,11 +527,14 @@ class SVN(Source):
                               to C{baseURL} and the result handed to
                               the SVN command.
 
+        @type  username: string
         @param username: username to pass to svn's --username
-        @param password: username to pass to svn's --password
+
+        @type  password: string
+        @param password: password to pass to svn's --password
         """
 
-        if not kwargs.has_key('workdir') and directory is not None:
+        if not 'workdir' in kwargs and directory is not None:
             # deal with old configs
             warn("Please use workdir=, not directory=", DeprecationWarning)
             kwargs['workdir'] = directory
@@ -408,7 +548,7 @@ class SVN(Source):
         self.keep_on_purge = keep_on_purge
         self.ignore_ignores = ignore_ignores
         self.always_purge = always_purge
-	self.depth = depth
+        self.depth = depth
 
         Source.__init__(self, **kwargs)
         self.addFactoryArguments(svnurl=svnurl,
@@ -421,12 +561,11 @@ class SVN(Source):
                                  keep_on_purge=keep_on_purge,
                                  ignore_ignores=ignore_ignores,
                                  always_purge=always_purge,
-				 depth=depth,
+                                 depth=depth,
                                  )
 
-        if not svnurl and not baseURL:
-            raise ValueError("you must use exactly one of svnurl and baseURL")
-
+        if svnurl and baseURL:
+            raise ValueError("you must use either svnurl OR baseURL")
 
     def computeSourceRevision(self, changes):
         if not changes or None in [c.revision for c in changes]:
@@ -434,11 +573,11 @@ class SVN(Source):
         lastChange = max([int(c.revision) for c in changes])
         return lastChange
 
-    def startVC(self, branch, revision, patch):
+    def checkCompatibility(self):
+        ''' Handle compatibility between old slaves/svn clients '''
 
-        # handle old slaves
-        warnings = []
         slavever = self.slaveVersion("svn", "old")
+
         if not slavever:
             m = "slave does not have the 'svn' command"
             raise BuildSlaveTooOldError(m)
@@ -449,7 +588,7 @@ class SVN(Source):
             # the last build used, so if we're using a non-default branch and
             # either 'update' or 'copy' modes, it is safer to refuse to
             # build, and tell the user they need to upgrade the buildslave.
-            if (branch != self.branch
+            if (self.args['branch'] != self.branch
                 and self.args['mode'] in ("update", "copy")):
                 m = ("This buildslave (%s) does not know about multiple "
                      "branches, and using mode=%s would probably build the "
@@ -459,66 +598,64 @@ class SVN(Source):
                                                    self.args['mode']))
                 raise BuildSlaveTooOldError(m)
 
-        if slavever == "old":
-            # 0.5.0 compatibility
-            if self.args['mode'] in ("clobber", "copy"):
-                # TODO: use some shell commands to make up for the
-                # deficiency, by blowing away the old directory first (thus
-                # forcing a full checkout)
-                warnings.append("WARNING: this slave can only do SVN updates"
-                                ", not mode=%s\n" % self.args['mode'])
-                log.msg("WARNING: this slave only does mode=update")
-            if self.args['mode'] == "export":
-                raise BuildSlaveTooOldError("old slave does not have "
-                                            "mode=export")
-            self.args['directory'] = self.args['workdir']
-            if revision is not None:
-                # 0.5.0 can only do HEAD. We have no way of knowing whether
-                # the requested revision is HEAD or not, and for
-                # slowly-changing trees this will probably do the right
-                # thing, so let it pass with a warning
-                m = ("WARNING: old slave can only update to HEAD, not "
-                     "revision=%s" % revision)
-                log.msg(m)
-                warnings.append(m + "\n")
-            revision = "HEAD" # interprets this key differently
-            if patch:
-                raise BuildSlaveTooOldError("old slave can't do patch")
+        if (self.depth is not None) and self.slaveVersionIsOlderThan("svn","2.9"):
+            m = ("This buildslave (%s) does not support svn depth "
+                    "arguments.  Refusing to build. "
+                    "Please upgrade the buildslave." % (self.build.slavename))
+            raise BuildSlaveTooOldError(m)
 
+        if (self.username is not None or self.password is not None) \
+        and self.slaveVersionIsOlderThan("svn", "2.8"):
+            m = ("This buildslave (%s) does not support svn usernames "
+                 "and passwords.  "
+                 "Refusing to build. Please upgrade the buildslave to "
+                 "buildbot-0.7.10 or newer." % (self.build.slavename,))
+            raise BuildSlaveTooOldError(m)
+
+    def getSvnUrl(self, branch, revision, patch):
+        ''' Compute the svn url that will be passed to the svn remote command '''
         if self.svnurl:
-            assert not branch # we need baseURL= to use branches
-            self.args['svnurl'] = self.svnurl
+            return self.computeRepositoryURL(self.svnurl)
         else:
-            self.args['svnurl'] = self.baseURL + branch
+            if branch is None:
+                m = ("The SVN source step belonging to builder '%s' does not know "
+                     "which branch to work with. This means that the change source "
+                     "did not specify a branch and that defaultBranch is None." \
+                     % self.build.builder.name)
+                raise RuntimeError(m)
+
+            computed = self.computeRepositoryURL(self.baseURL)
+
+            if self.branch_placeholder in self.baseURL:
+                return computed.replace(self.branch_placeholder, branch)
+            else:
+                return computed + branch
+
+    def startVC(self, branch, revision, patch):
+        warnings = []
+
+        self.checkCompatibility()
+
+        self.args['svnurl'] = self.getSvnUrl(branch, revision, patch)
         self.args['revision'] = revision
         self.args['patch'] = patch
+        self.args['always_purge'] = self.always_purge
 
-	#Set up depth if specified
-	if self.depth is not None:
-		if self.slaveVersionIsOlderThan("svn","2.9"):
-			m = ("This buildslave (%s) does not support svn depth "
-			     "arguments.  "
-			     "Refusing to build. "
-			     "Please upgrade the buildslave." % (self.build.slavename))
-                	raise BuildSlaveTooOldError(m)
-		else: 
-			self.args['depth'] = self.depth
+        #Set up depth if specified
+        if self.depth is not None:
+            self.args['depth'] = self.depth
 
-        if self.username is not None or self.password is not None:
-            if self.slaveVersionIsOlderThan("svn", "2.8"):
-                m = ("This buildslave (%s) does not support svn usernames "
-                     "and passwords.  "
-                     "Refusing to build. Please upgrade the buildslave to "
-                     "buildbot-0.7.10 or newer." % (self.build.slavename,))
-                raise BuildSlaveTooOldError(m)
-            if self.username is not None: self.args['username'] = self.username
-            if self.password is not None: self.args['password'] = self.password
+        if self.username is not None:
+            self.args['username'] = self.username
+        if self.password is not None:
+            self.args['password'] = self.password
 
         if self.extra_args is not None:
             self.args['extra_args'] = self.extra_args
 
         revstuff = []
-        if branch is not None and branch != self.branch:
+        #revstuff.append(self.args['svnurl'])
+        if self.args['svnurl'].find('trunk') == -1:
             revstuff.append("[branch]")
         if revision is not None:
             revstuff.append("r%s" % revision)
@@ -574,7 +711,7 @@ class Darcs(Source):
                                  )
         assert self.args['mode'] != "export", \
                "Darcs does not have an 'export' mode"
-        if (not repourl and not baseURL) or (repourl and baseURL):
+        if repourl and baseURL:
             raise ValueError("you must provide exactly one of repourl and"
                              " baseURL")
 
@@ -607,9 +744,9 @@ class Darcs(Source):
 
         if self.repourl:
             assert not branch # we need baseURL= to use branches
-            self.args['repourl'] = self.repourl
+            self.args['repourl'] = self.computeRepositoryURL(self.repourl)
         else:
-            self.args['repourl'] = self.baseURL + branch
+            self.args['repourl'] = self.computeRepositoryURL(self.baseURL) + branch
         self.args['revision'] = revision
         self.args['patch'] = patch
 
@@ -628,10 +765,13 @@ class Git(Source):
 
     name = "git"
 
-    def __init__(self, repourl,
+    def __init__(self, repourl=None,
                  branch="master",
                  submodules=False,
                  ignore_ignores=None,
+                 reference=None,
+                 shallow=False,
+                 progress=False,
                  **kwargs):
         """
         @type  repourl: string
@@ -646,17 +786,34 @@ class Git(Source):
         @param submodules: Whether or not to update (and initialize)
                        git submodules.
 
+        @type  reference: string
+        @param reference: The path to a reference repository to obtain
+                          objects from, if any.
+
+        @type  shallow: boolean
+        @param shallow: Use a shallow or clone, if possible
+
+        @type  progress: boolean
+        @param progress: Pass the --progress option when fetching. This
+                         can solve long fetches getting killed due to
+                         lack of output, but requires Git 1.7.2+.
         """
         Source.__init__(self, **kwargs)
+        self.repourl = repourl
         self.addFactoryArguments(repourl=repourl,
                                  branch=branch,
                                  submodules=submodules,
                                  ignore_ignores=ignore_ignores,
+                                 reference=reference,
+                                 shallow=shallow,
+                                 progress=progress,
                                  )
-        self.args.update({'repourl': repourl,
-                          'branch': branch,
+        self.args.update({'branch': branch,
                           'submodules': submodules,
                           'ignore_ignores': ignore_ignores,
+                          'reference': reference,
+                          'shallow': shallow,
+                          'progress': progress,
                           })
 
     def computeSourceRevision(self, changes):
@@ -668,6 +825,7 @@ class Git(Source):
         if branch is not None:
             self.args['branch'] = branch
 
+        self.args['repourl'] = self.computeRepositoryURL(self.repourl)
         self.args['revision'] = revision
         self.args['patch'] = patch
         slavever = self.slaveVersion("git")
@@ -678,172 +836,6 @@ class Git(Source):
         self.startCommand(cmd)
 
 
-class Arch(Source):
-    """Check out a source tree from an Arch repository named 'archive'
-    available at 'url'. 'version' specifies which version number (development
-    line) will be used for the checkout: this is mostly equivalent to a
-    branch name. This version uses the 'tla' tool to do the checkout, to use
-    'baz' see L{Bazaar} instead.
-    """
-
-    name = "arch"
-    # TODO: slaves >0.6.6 will accept args['build-config'], so use it
-
-    def __init__(self, url, version, archive=None, **kwargs):
-        """
-        @type  url: string
-        @param url: the Arch coordinates of the repository. This is
-                    typically an http:// URL, but could also be the absolute
-                    pathname of a local directory instead.
-
-        @type  version: string
-        @param version: the category--branch--version to check out. This is
-                        the default branch. If a build specifies a different
-                        branch, it will be used instead of this.
-
-        @type  archive: string
-        @param archive: The archive name. If provided, it must match the one
-                        that comes from the repository. If not, the
-                        repository's default will be used.
-        """
-        self.branch = version
-        Source.__init__(self, **kwargs)
-        self.addFactoryArguments(url=url,
-                                 version=version,
-                                 archive=archive,
-                                 )
-        self.args.update({'url': url,
-                          'archive': archive,
-                          })
-
-    def computeSourceRevision(self, changes):
-        # in Arch, fully-qualified revision numbers look like:
-        #  arch@buildbot.sourceforge.net--2004/buildbot--dev--0--patch-104
-        # For any given builder, all of this is fixed except the patch-104.
-        # The Change might have any part of the fully-qualified string, so we
-        # just look for the last part. We return the "patch-NN" string.
-        if not changes:
-            return None
-        lastChange = None
-        for c in changes:
-            if not c.revision:
-                continue
-            if c.revision.endswith("--base-0"):
-                rev = 0
-            else:
-                i = c.revision.rindex("patch")
-                rev = int(c.revision[i+len("patch-"):])
-            lastChange = max(lastChange, rev)
-        if lastChange is None:
-            return None
-        if lastChange == 0:
-            return "base-0"
-        return "patch-%d" % lastChange
-
-    def checkSlaveVersion(self, cmd, branch):
-        warnings = []
-        slavever = self.slaveVersion(cmd)
-        if not slavever:
-            m = "slave is too old, does not know about %s" % cmd
-            raise BuildSlaveTooOldError(m)
-
-        # slave 1.28 and later understand 'revision'
-        if self.slaveVersionIsOlderThan(cmd, "1.28"):
-            if not self.alwaysUseLatest:
-                # we don't know whether our requested revision is the latest
-                # or not. If the tree does not change very quickly, this will
-                # probably build the right thing, so emit a warning rather
-                # than refuse to build at all
-                m = "WARNING, buildslave is too old to use a revision"
-                log.msg(m)
-                warnings.append(m + "\n")
-
-        if self.slaveVersionIsOlderThan(cmd, "1.39"):
-            # the slave doesn't know to avoid re-using the same sourcedir
-            # when the branch changes. We have no way of knowing which branch
-            # the last build used, so if we're using a non-default branch and
-            # either 'update' or 'copy' modes, it is safer to refuse to
-            # build, and tell the user they need to upgrade the buildslave.
-            if (branch != self.branch
-                and self.args['mode'] in ("update", "copy")):
-                m = ("This buildslave (%s) does not know about multiple "
-                     "branches, and using mode=%s would probably build the "
-                     "wrong tree. "
-                     "Refusing to build. Please upgrade the buildslave to "
-                     "buildbot-0.7.0 or newer." % (self.build.slavename,
-                                                   self.args['mode']))
-                log.msg(m)
-                raise BuildSlaveTooOldError(m)
-
-        return warnings
-
-    def startVC(self, branch, revision, patch):
-        self.args['version'] = branch
-        self.args['revision'] = revision
-        self.args['patch'] = patch
-        warnings = self.checkSlaveVersion("arch", branch)
-
-        revstuff = []
-        if branch is not None and branch != self.branch:
-            revstuff.append("[branch]")
-        if revision is not None:
-            revstuff.append("patch%s" % revision)
-        self.description.extend(revstuff)
-        self.descriptionDone.extend(revstuff)
-
-        cmd = LoggedRemoteCommand("arch", self.args)
-        self.startCommand(cmd, warnings)
-
-
-class Bazaar(Arch):
-    """Bazaar is an alternative client for Arch repositories. baz is mostly
-    compatible with tla, but archive registration is slightly different."""
-
-    # TODO: slaves >0.6.6 will accept args['build-config'], so use it
-
-    def __init__(self, url, version, archive, **kwargs):
-        """
-        @type  url: string
-        @param url: the Arch coordinates of the repository. This is
-                    typically an http:// URL, but could also be the absolute
-                    pathname of a local directory instead.
-
-        @type  version: string
-        @param version: the category--branch--version to check out
-
-        @type  archive: string
-        @param archive: The archive name (required). This must always match
-                        the one that comes from the repository, otherwise the
-                        buildslave will attempt to get sources from the wrong
-                        archive.
-        """
-        self.branch = version
-        Source.__init__(self, **kwargs)
-        self.addFactoryArguments(url=url,
-                                 version=version,
-                                 archive=archive,
-                                 )
-        self.args.update({'url': url,
-                          'archive': archive,
-                          })
-
-    def startVC(self, branch, revision, patch):
-        self.args['version'] = branch
-        self.args['revision'] = revision
-        self.args['patch'] = patch
-        warnings = self.checkSlaveVersion("bazaar", branch)
-
-        revstuff = []
-        if branch is not None and branch != self.branch:
-            revstuff.append("[branch]")
-        if revision is not None:
-            revstuff.append("patch%s" % revision)
-        self.description.extend(revstuff)
-        self.descriptionDone.extend(revstuff)
-
-        cmd = LoggedRemoteCommand("bazaar", self.args)
-        self.startCommand(cmd, warnings)
-
 class Bzr(Source):
     """Check out a source tree from a bzr (Bazaar) repository at 'repourl'.
 
@@ -892,7 +884,7 @@ class Bzr(Source):
                                  forceSharedRepo=forceSharedRepo
                                  )
         self.args.update({'forceSharedRepo': forceSharedRepo})
-        if (not repourl and not baseURL) or (repourl and baseURL):
+        if repourl and baseURL:
             raise ValueError("you must provide exactly one of repourl and"
                              " baseURL")
 
@@ -910,9 +902,9 @@ class Bzr(Source):
 
         if self.repourl:
             assert not branch # we need baseURL= to use branches
-            self.args['repourl'] = self.repourl
+            self.args['repourl'] = self.computeRepositoryURL(self.repourl)
         else:
-            self.args['repourl'] = self.baseURL + branch
+            self.args['repourl'] = self.computeRepositoryURL(self.baseURL) + branch
         self.args['revision'] = revision
         self.args['patch'] = patch
 
@@ -980,7 +972,7 @@ class Mercurial(Source):
                                  branchType=branchType,
                                  clobberOnBranchChange=clobberOnBranchChange,
                                  )
-        if (not repourl and not baseURL) or (repourl and baseURL):
+        if repourl and baseURL:
             raise ValueError("you must provide exactly one of repourl and"
                              " baseURL")
 
@@ -993,11 +985,11 @@ class Mercurial(Source):
         if self.repourl:
             # we need baseURL= to use dirname branches
             assert self.branchType == 'inrepo' or not branch
-            self.args['repourl'] = self.repourl
+            self.args['repourl'] = self.computeRepositoryURL(self.repourl)
             if branch:
                 self.args['branch'] = branch
         else:
-            self.args['repourl'] = self.baseURL + (branch or '')
+            self.args['repourl'] = self.computeRepositoryURL(self.baseURL) + (branch or '')
         self.args['revision'] = revision
         self.args['patch'] = patch
         self.args['clobberOnBranchChange'] = self.clobberOnBranchChange
@@ -1030,8 +1022,8 @@ class P4(Source):
     """ P4 is a class for accessing perforce revision control"""
     name = "p4"
 
-    def __init__(self, p4base, defaultBranch=None, p4port=None, p4user=None,
-                 p4passwd=None, p4extra_views=[],
+    def __init__(self, p4base=None, defaultBranch=None, p4port=None, p4user=None,
+                 p4passwd=None, p4extra_views=[], p4line_end='local',
                  p4client='buildbot_%(slave)s_%(builder)s', **kwargs):
         """
         @type  p4base: string
@@ -1059,10 +1051,14 @@ class P4(Source):
         @param p4extra_views: Extra views to be added to
                               the client that is being used.
 
+        @type  p4line_end: string
+        @param p4line_end: value of the LineEnd client specification property
+
         @type  p4client: string
         @param p4client: The perforce client to use for this buildslave.
         """
 
+        self.p4base = p4base
         self.branch = defaultBranch
         Source.__init__(self, **kwargs)
         self.addFactoryArguments(p4base=p4base,
@@ -1071,17 +1067,19 @@ class P4(Source):
                                  p4user=p4user,
                                  p4passwd=p4passwd,
                                  p4extra_views=p4extra_views,
+                                 p4line_end=p4line_end,
                                  p4client=p4client,
                                  )
         self.args['p4port'] = p4port
         self.args['p4user'] = p4user
         self.args['p4passwd'] = p4passwd
-        self.args['p4base'] = p4base
         self.args['p4extra_views'] = p4extra_views
+        self.args['p4line_end'] = p4line_end
         self.p4client = p4client
 
     def setBuild(self, build):
         Source.setBuild(self, build)
+        self.args['p4base'] = self.computeRepositoryURL(self.p4base)
         self.args['p4client'] = self.p4client % {
             'slave': build.slavename,
             'builder': build.builder.name,
@@ -1148,39 +1146,3 @@ class P4Sync(Source):
         assert slavever, "slave is too old, does not know about p4"
         cmd = LoggedRemoteCommand("p4sync", self.args)
         self.startCommand(cmd)
-
-class Monotone(Source):
-    """Check out a revision from a monotone server at 'server_addr',
-    branch 'branch'.  'revision' specifies which revision id to check
-    out.
-
-    This step will first create a local database, if necessary, and then pull
-    the contents of the server into the database.  Then it will do the
-    checkout/update from this database."""
-
-    name = "monotone"
-
-    def __init__(self, server_addr, branch, db_path="monotone.db",
-                 monotone="monotone",
-                 **kwargs):
-        Source.__init__(self, **kwargs)
-        self.addFactoryArguments(server_addr=server_addr,
-                                 branch=branch,
-                                 db_path=db_path,
-                                 monotone=monotone,
-                                 )
-        self.args.update({"server_addr": server_addr,
-                          "branch": branch,
-                          "db_path": db_path,
-                          "monotone": monotone})
-
-    def computeSourceRevision(self, changes):
-        if not changes:
-            return None
-        return changes[-1].revision
-
-    def startVC(self):
-        slavever = self.slaveVersion("monotone")
-        assert slavever, "slave is too old, does not know about monotone"
-        cmd = LoggedRemoteCommand("monotone", self.args)
-        self.startCommand(cmd)
