@@ -1,12 +1,17 @@
 # -*- test-case-name: buildbot.test.test_transfer -*-
 
-import os.path
+import os.path, tarfile, tempfile
+try:
+    from cStringIO import StringIO
+except ImportError:
+    from StringIO import StringIO
 from twisted.internet import reactor
 from twisted.spread import pb
 from twisted.python import log
 from buildbot.process.buildstep import RemoteCommand, BuildStep
-from buildbot.process.buildstep import SUCCESS, FAILURE
+from buildbot.process.buildstep import SUCCESS, FAILURE, SKIPPED
 from buildbot.interfaces import BuildSlaveTooOldError
+from buildbot.util import json
 
 
 class _FileWriter(pb.Referenceable):
@@ -22,19 +27,20 @@ class _FileWriter(pb.Referenceable):
             os.makedirs(dirname)
 
         self.destfile = destfile
-        self.fp = open(destfile, "wb")
+        fd, self.tmpname = tempfile.mkstemp(dir=dirname)
+        self.fp = os.fdopen(fd, 'wb')
         if mode is not None:
             os.chmod(destfile, mode)
-	self.remaining = maxsize
+        self.remaining = maxsize
 
     def remote_write(self, data):
-	"""
-	Called from remote slave to write L{data} to L{fp} within boundaries
-	of L{maxsize}
+        """
+        Called from remote slave to write L{data} to L{fp} within boundaries
+        of L{maxsize}
 
-	@type  data: C{string}
-	@param data: String of data to write
-	"""
+        @type  data: C{string}
+        @param data: String of data to write
+        """
         if self.remaining is not None:
             if len(data) > self.remaining:
                 data = data[:self.remaining]
@@ -49,6 +55,8 @@ class _FileWriter(pb.Referenceable):
         """
         self.fp.close()
         self.fp = None
+        os.rename(self.tmpname, self.destfile)
+        self.tmpname = None
 
     def __del__(self):
         # unclean shutdown, the file is probably truncated, so delete it
@@ -57,74 +65,84 @@ class _FileWriter(pb.Referenceable):
         if fp:
             fp.close()
             os.unlink(self.destfile)
+            if self.tmpname and os.path.exists(self.tmpname):
+                os.unlink(self.tmpname)
 
 
-class _DirectoryWriter(pb.Referenceable):
+def _extractall(self, path=".", members=None):
+    """Fallback extractall method for TarFile, in case it doesn't have its own."""
+
+    import copy
+
+    directories = []
+
+    if members is None:
+        members = self
+
+    for tarinfo in members:
+        if tarinfo.isdir():
+            # Extract directories with a safe mode.
+            directories.append(tarinfo)
+            tarinfo = copy.copy(tarinfo)
+            tarinfo.mode = 0700
+        self.extract(tarinfo, path)
+
+    # Reverse sort directories.
+    directories.sort(lambda a, b: cmp(a.name, b.name))
+    directories.reverse()
+
+    # Set correct owner, mtime and filemode on directories.
+    for tarinfo in directories:
+        dirpath = os.path.join(path, tarinfo.name)
+        try:
+            self.chown(tarinfo, dirpath)
+            self.utime(tarinfo, dirpath)
+            self.chmod(tarinfo, dirpath)
+        except tarfile.ExtractError, e:
+            if self.errorlevel > 1:
+                raise
+            else:
+                self._dbg(1, "tarfile: %s" % e)
+
+class _DirectoryWriter(_FileWriter):
     """
-    Helper class that acts as a directory-object with write access
+    A DirectoryWriter is implemented as a FileWriter, with an added post-processing
+    step to unpack the archive, once the transfer has completed.
     """
 
-    def __init__(self, destroot, maxsize, mode):
+    def __init__(self, destroot, maxsize, compress, mode):
         self.destroot = destroot
-	# Create missing directories.
-        self.destroot = os.path.abspath(self.destroot)
-        if not os.path.exists(self.destroot):
-            os.makedirs(self.destroot)
-	
-	self.fp = None
-	self.mode = mode
-	self.maxsize = maxsize
+        self.compress = compress
 
-    def remote_createdir(self, dirname):
-	# This function is needed to transfer empty directories.
-	dirname = os.path.join(self.destroot, dirname)
-	dirname = os.path.abspath(dirname)
-	if not os.path.exists(dirname):
-            os.makedirs(dirname)
+        self.fd, self.tarname = tempfile.mkstemp()
+        os.close(self.fd)
 
-    def remote_open(self, destfile):
-	# Create missing directories.
-	destfile = os.path.join(self.destroot, destfile)
-        destfile = os.path.abspath(destfile)
-        dirname = os.path.dirname(destfile)
-        if not os.path.exists(dirname):
-            os.makedirs(dirname)
+        _FileWriter.__init__(self, self.tarname, maxsize, mode)
 
-        self.fp = open(destfile, "wb")
-        if self.mode is not None:
-            os.chmod(destfile, self.mode)
-	self.remaining = self.maxsize
-
-    def remote_write(self, data):
-	"""
-	Called from remote slave to write L{data} to L{fp} within boundaries
-	of L{maxsize}
-
-	@type  data: C{string}
-	@param data: String of data to write
-	"""
-        if self.remaining is not None:
-            if len(data) > self.remaining:
-                data = data[:self.remaining]
-            self.fp.write(data)
-            self.remaining = self.remaining - len(data)
-        else:
-            self.fp.write(data)
-
-    def remote_close(self):
+    def remote_unpack(self):
         """
         Called by remote slave to state that no more data will be transfered
         """
-	if self.fp:
-	    self.fp.close()
-    	    self.fp = None
+        # Make sure remote_close is called, otherwise atomic rename wont happen
+        self.remote_close()
 
-    def __del__(self):
-        # unclean shutdown, the file is probably truncated, so delete it
-        # altogether rather than deliver a corrupted file
-        fp = getattr(self, "fp", None)
-        if fp:
-            fp.close()
+        # Map configured compression to a TarFile setting
+        if self.compress == 'bz2':
+            mode='r|bz2'
+        elif self.compress == 'gz':
+            mode='r|gz'
+        else:
+            mode = 'r'
+
+        # Support old python
+        if not hasattr(tarfile.TarFile, 'extractall'):
+            tarfile.TarFile.extractall = _extractall
+
+        # Unpack archive and clean up after self
+        archive = tarfile.open(name=self.tarname, mode=mode)
+        archive.extractall(path=self.destroot)
+        archive.close()
+        os.remove(self.tarname)
 
 
 class StatusRemoteCommand(RemoteCommand):
@@ -160,6 +178,19 @@ class _TransferBuildStep(BuildStep):
             workdir = self.workdir
         return properties.render(workdir)
 
+    def finished(self, result):
+        # Subclasses may choose to skip a transfer. In those cases, self.cmd
+        # will be None, and we should just let BuildStep.finished() handle
+        # the rest
+        if result == SKIPPED:
+            return BuildStep.finished(self, SKIPPED)
+        if self.cmd.stderr != '':
+            self.addCompleteLog('stderr', self.cmd.stderr)
+
+        if self.cmd.rc is None or self.cmd.rc == 0:
+            return BuildStep.finished(self, SUCCESS)
+        return BuildStep.finished(self, FAILURE)
+
 
 class FileUpload(_TransferBuildStep):
     """
@@ -237,14 +268,6 @@ class FileUpload(_TransferBuildStep):
         d = self.runCommand(self.cmd)
         d.addCallback(self.finished).addErrback(self.failed)
 
-    def finished(self, result):
-        if self.cmd.stderr != '':
-            self.addCompleteLog('stderr', self.cmd.stderr)
-
-        if self.cmd.rc is None or self.cmd.rc == 0:
-            return BuildStep.finished(self, SUCCESS)
-        return BuildStep.finished(self, FAILURE)
-
 
 class DirectoryUpload(BuildStep):
     """
@@ -256,26 +279,25 @@ class DirectoryUpload(BuildStep):
     - ['masterdest'] name of destination directory at master
     - ['workdir']    string with slave working directory relative to builder
                      base dir, default 'build'
-    - ['maxsize']    maximum size of each file, default None (=unlimited)
+    - ['maxsize']    maximum size of the compressed tarfile containing the
+                     whole directory
     - ['blocksize']  maximum size of each block being transfered
-    - ['mode']       file access mode for the resulting master-side file.
-                     The default (=None) is to leave it up to the umask of
-                     the buildmaster process.
+    - ['compress']   compression type to use: one of [None, 'gz', 'bz2']
 
     """
 
     name = 'upload'
 
     def __init__(self, slavesrc, masterdest,
-                 workdir="build", maxsize=None, blocksize=16*1024, mode=None,
-                 **buildstep_kwargs):
+                 workdir="build", maxsize=None, blocksize=16*1024,
+                 compress=None, **buildstep_kwargs):
         BuildStep.__init__(self, **buildstep_kwargs)
         self.addFactoryArguments(slavesrc=slavesrc,
                                  masterdest=masterdest,
                                  workdir=workdir,
                                  maxsize=maxsize,
                                  blocksize=blocksize,
-                                 mode=mode,
+                                 compress=compress,
                                  )
 
         self.slavesrc = slavesrc
@@ -283,8 +305,8 @@ class DirectoryUpload(BuildStep):
         self.workdir = workdir
         self.maxsize = maxsize
         self.blocksize = blocksize
-        assert isinstance(mode, (int, type(None)))
-        self.mode = mode
+        assert compress in (None, 'gz', 'bz2')
+        self.compress = compress
 
     def start(self):
         version = self.slaveVersion("uploadDirectory")
@@ -305,9 +327,9 @@ class DirectoryUpload(BuildStep):
                 % (source, masterdest))
 
         self.step_status.setText(['uploading', os.path.basename(source)])
-	
+        
         # we use maxsize to limit the amount of data on both sides
-        dirWriter = _DirectoryWriter(masterdest, self.maxsize, self.mode)
+        dirWriter = _DirectoryWriter(masterdest, self.maxsize, self.compress, 0600)
 
         # default arguments
         args = {
@@ -316,6 +338,7 @@ class DirectoryUpload(BuildStep):
             'writer': dirWriter,
             'maxsize': self.maxsize,
             'blocksize': self.blocksize,
+            'compress': self.compress
             }
 
         self.cmd = StatusRemoteCommand('uploadDirectory', args)
@@ -323,6 +346,11 @@ class DirectoryUpload(BuildStep):
         d.addCallback(self.finished).addErrback(self.failed)
 
     def finished(self, result):
+        # Subclasses may choose to skip a transfer. In those cases, self.cmd
+        # will be None, and we should just let BuildStep.finished() handle
+        # the rest
+        if result == SKIPPED:
+            return BuildStep.finished(self, SKIPPED)
         if self.cmd.stderr != '':
             self.addCompleteLog('stderr', self.cmd.stderr)
 
@@ -342,15 +370,15 @@ class _FileReader(pb.Referenceable):
         self.fp = fp
 
     def remote_read(self, maxlength):
-	"""
-	Called from remote slave to read at most L{maxlength} bytes of data
+        """
+        Called from remote slave to read at most L{maxlength} bytes of data
 
-	@type  maxlength: C{integer}
-	@param maxlength: Maximum number of data bytes that can be returned
+        @type  maxlength: C{integer}
+        @param maxlength: Maximum number of data bytes that can be returned
 
         @return: Data read from L{fp}
         @rtype: C{string} of bytes read from file
-	"""
+        """
         if self.fp is None:
             return ''
 
@@ -455,11 +483,120 @@ class FileDownload(_TransferBuildStep):
         d = self.runCommand(self.cmd)
         d.addCallback(self.finished).addErrback(self.failed)
 
-    def finished(self, result):
-        if self.cmd.stderr != '':
-            self.addCompleteLog('stderr', self.cmd.stderr)
+class StringDownload(_TransferBuildStep):
+    """
+    Download the first 'maxsize' bytes of a string, from the buildmaster to the
+    buildslave. Set the mode of the file
 
-        if self.cmd.rc is None or self.cmd.rc == 0:
-            return BuildStep.finished(self, SUCCESS)
-        return BuildStep.finished(self, FAILURE)
+    Arguments::
 
+     ['s']         string to transfer
+     ['slavedest'] filename of destination file at slave
+     ['workdir']   string with slave working directory relative to builder
+                   base dir, default 'build'
+     ['maxsize']   maximum size of the file, default None (=unlimited)
+     ['blocksize'] maximum size of each block being transfered
+     ['mode']      use this to set the access permissions of the resulting
+                   buildslave-side file. This is traditionally an octal
+                   integer, like 0644 to be world-readable (but not
+                   world-writable), or 0600 to only be readable by
+                   the buildslave account, or 0755 to be world-executable.
+                   The default (=None) is to leave it up to the umask of
+                   the buildslave process.
+    """
+    name = 'string_download'
+
+    def __init__(self, s, slavedest,
+                 workdir=None, maxsize=None, blocksize=16*1024, mode=None,
+                 **buildstep_kwargs):
+        BuildStep.__init__(self, **buildstep_kwargs)
+        self.addFactoryArguments(s=s,
+                                 slavedest=slavedest,
+                                 workdir=workdir,
+                                 maxsize=maxsize,
+                                 blocksize=blocksize,
+                                 mode=mode,
+                                 )
+
+        self.s = s
+        self.slavedest = slavedest
+        self.workdir = workdir
+        self.maxsize = maxsize
+        self.blocksize = blocksize
+        assert isinstance(mode, (int, type(None)))
+        self.mode = mode
+
+    def start(self):
+        properties = self.build.getProperties()
+
+        version = self.slaveVersion("downloadFile")
+        if not version:
+            m = "slave is too old, does not know about downloadFile"
+            raise BuildSlaveTooOldError(m)
+
+        # we are currently in the buildmaster's basedir, so any non-absolute
+        # paths will be interpreted relative to that
+        slavedest = properties.render(self.slavedest)
+        log.msg("StringDownload started, from master to slave %r" % slavedest)
+
+        self.step_status.setText(['downloading', "to",
+                                  os.path.basename(slavedest)])
+
+        # setup structures for reading the file
+        fp = StringIO(self.s)
+        fileReader = _FileReader(fp)
+
+        # default arguments
+        args = {
+            'slavedest': slavedest,
+            'maxsize': self.maxsize,
+            'reader': fileReader,
+            'blocksize': self.blocksize,
+            'workdir': self._getWorkdir(),
+            'mode': self.mode,
+            }
+
+        self.cmd = StatusRemoteCommand('downloadFile', args)
+        d = self.runCommand(self.cmd)
+        d.addCallback(self.finished).addErrback(self.failed)
+
+class JSONStringDownload(StringDownload):
+    """
+    Encode object o as a json string and save it on the buildslave
+
+    Arguments::
+
+     ['o']         object to encode and transfer
+    """
+    name = "json_download"
+    def __init__(self, o, slavedest, **buildstep_kwargs):
+        if 's' in buildstep_kwargs:
+            del buildstep_kwargs['s']
+        s = json.dumps(o)
+        StringDownload.__init__(self, s=s, slavedest=slavedest, **buildstep_kwargs)
+        self.addFactoryArguments(o=o)
+
+class JSONPropertiesDownload(StringDownload):
+    """
+    Download the current build properties as a json string and save it on the
+    buildslave
+    """
+    name = "json_properties_download"
+    def __init__(self, slavedest, **buildstep_kwargs):
+        self.super_class = StringDownload
+        if 's' in buildstep_kwargs:
+            del buildstep_kwargs['s']
+        StringDownload.__init__(self, s=None, slavedest=slavedest, **buildstep_kwargs)
+
+    def start(self):
+        properties = self.build.getProperties()
+        props = {}
+        for key, value, source in properties.asList():
+            props[key] = value
+
+        self.s = json.dumps(dict(
+                        properties=props,
+                        sourcestamp=self.build.getSourceStamp().asDict(),
+                    ),
+                )
+        return self.super_class.start(self)
